"""
Test all methods which return collections of gene IDs that aren't converting
from some other type of name or ID.

TODO: Implement tests for EnsemblRelease.gene_ids
"""
from __future__ import absolute_import

from nose.tools import assert_raises, ok_
from pyensembl import ensembl_grch38, ensembl77

from .common import test_ensembl_releases


def test_gene_ids_grch38_hla_a():
    # chr6:29,945,884  is a position for HLA-A
    # Gene ID = ENSG00000206503
    # based on:
    # http://useast.ensembl.org/Homo_sapiens/Gene/
    # Summary?db=core;g=ENSG00000206503;r=6:29941260-29945884
    ids = ensembl_grch38.gene_ids_at_locus(6, 29945884)
    expected = "ENSG00000206503"
    assert ids == ["ENSG00000206503"], \
        "Expected HLA-A, gene ID = %s, got: %s" % (expected, ids)

def test_gene_ids_of_gene_name_hla_grch38():
    hla_a_gene_ids = ensembl_grch38.gene_ids_of_gene_name("HLA-A")
    assert 'ENSG00000206503' in hla_a_gene_ids, hla_a_gene_ids

    hla_b_gene_ids = ensembl_grch38.gene_ids_of_gene_name("HLA-B")
    assert 'ENSG00000234745' in hla_b_gene_ids, hla_b_gene_ids

    hla_c_gene_ids = ensembl_grch38.gene_ids_of_gene_name("HLA-C")
    assert 'ENSG00000204525' in hla_c_gene_ids, hla_c_gene_ids

def test_gene_id_of_protein_id_release77():
    gene_id = ensembl77.gene_id_of_protein_id("ENSP00000485677")
    ok_('ENSG00000279634', gene_id)

def test_gene_id_of_invalid_name():
    with assert_raises(Exception):
        ensembl_grch38.gene_ids_of_gene_name(
            "A wonderous pony sees through your soul")

@test_ensembl_releases()
def test_gene_ids_on_contig(ensembl):
    gene_ids_chr17 = ensembl.gene_ids(contig=17)
    # gene ID of TP53
    tp53 = "ENSG00000141510"
    assert tp53 in gene_ids_chr17, \
        "Missing %s from %s on chr17, example IDs: %s (total = %d)" % (
            tp53, ensembl, gene_ids_chr17[:5], len(gene_ids_chr17))

    # gene ID of SMAD4
    gene_ids_chr18 = ensembl.gene_ids(contig=18)
    smad4 = "ENSG00000141646"
    assert smad4 in gene_ids_chr18, \
        "Missing %s from %s on chr18, example result: %s (total = %d)" % (
            smad4, ensembl, gene_ids_chr18[:5], len(gene_ids_chr18))



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from .memory_cache import MemoryCache
from .download_cache import DownloadCache
from .ensembl_release import EnsemblRelease
from .ensembl_release_versions import check_release_number, MAX_ENSEMBL_RELEASE
from .genome import Genome
from .gene import Gene
from .gtf import GTF
from .locus import Locus
from .search import find_nearest_locus
from .sequence_data import SequenceData
from .species import (
    find_species_by_name,
    find_species_by_reference,
    which_reference,
    check_species_object,
    normalize_reference_name,
    normalize_species_name,
)
from .transcript import Transcript

_cache = {}

def cached_release(version, species="human"):
    """Cached construction of EnsemblRelease objects. It's desirable to reuse
    the same EnsemblRelease object since each one will store a lot of cached
    annotation data in-memory.
    """
    version = check_release_number(version)
    species = check_species_object(species)
    key = (version, species)
    if key not in _cache:
        ensembl = EnsemblRelease(version, species=species)
        _cache[key] = ensembl
    return _cache[key]

def genome_for_reference_name(reference_name):
    reference_name = normalize_reference_name(reference_name)
    species = find_species_by_reference(reference_name)
    (_, max_ensembl_release) = species.reference_assemblies[reference_name]
    return cached_release(max_ensembl_release, species=species)

ensembl_grch36 = ensembl54 = cached_release(54)  # last release for GRCh36/hg18
ensembl_grch37 = ensembl75 = cached_release(75)  # last release for GRCh37/hg19

ensembl77 = cached_release(77)
ensembl78 = cached_release(78)
ensembl79 = cached_release(79)
ensembl80 = cached_release(80)
ensembl81 = cached_release(81)
ensembl_grch38 = ensembl81  # most recent for GRCh38



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Ensembl/GENCODE biotype classifications, for more information:
http://useast.ensembl.org/Help/Faq?id=468

Definitions for GENCODE biotypes from:
http://www.gencodegenes.org/gencode_biotypes.html

(not all of these are necessarily used in Ensembl)

IG_C_gene
IG_D_gene
IG_J_gene
IG_V_gene
TR_C_gene
TR_J_gene
TR_V_gene
TR_D_gene
----------------
Immunoglobulin (Ig) variable chain and T-cell receptor (TcR) genes imported
or annotated according to the IMGT. IG_C_pseudogene

IG_J_pseudogene
IG_V_pseudogene
TR_V_pseudogene
TR_J_pseudogene
----------------
Inactivated immunoglobulin gene.

Mt_rRNA
Mt_tRNA
miRNA
misc_RNA
rRNA
snRNA
snoRNA
----------------
Non-coding RNA predicted using sequences from RFAM and miRBase

Mt_tRNA_pseudogene
tRNA_pseudogene
snoRNA_pseudogene
snRNA_pseudogene
scRNA_pseudogene
rRNA_pseudogene
misc_RNA_pseudogene
miRNA_pseudogene
Non-coding RNAs predicted to be pseudogenes by the Ensembl pipeline

TEC
----------------
To be Experimentally Confirmed. This is used for non-spliced EST clusters that
have polyA features. This category has been specifically created for the ENCODE
project to highlight regions that could indicate the presence of protein coding
genes that require experimental validation, either by 5' RACE or RT-PCR to
extend the transcripts, or by confirming expression of the putatively-encoded
peptide with specific antibodies.

nonsense_mediated_decay
----------------
If the coding sequence (following the appropriate reference) of a transcript
finishes >50bp from a downstream splice site then it is tagged as NMD.
If the variant does not cover the full reference coding sequence then it is
annotated as NMD if NMD is unavoidable i.e. no matter what the exon structure
of the missing portion is the transcript will be subject to NMD.

non_stop_decay
----------------
Transcripts that have polyA features (including signal) without a prior stop
codon in the CDS, i.e. a non-genomic polyA tail attached directly to the CDS
without 3' UTR. These transcripts are subject to degradation.


retained_intron
----------------
Alternatively spliced transcript believed to contain intronic sequence relative
to other, coding, variants. protein_coding Contains an open reading frame (ORF).

processed_transcript
----------------
Doesn't contain an ORF.

non_coding
----------------
Transcript which is known from the literature to not be protein coding.

ambiguous_orf
----------------
Transcript believed to be protein coding, but with more than one possible
open reading frame.

sense_intronic
----------------
Long non-coding transcript in introns of a coding gene that does not overlap any
exons.

sense_overlapping
----------------
Long non-coding transcript that contains a coding gene in its intron on the same
strand.

antisense
----------------
Has transcripts that overlap the genomic span (i.e. exon or introns) of a
protein-coding locus on the opposite strand.

known_ncrna
----------------

pseudogene
----------------
Have homology to proteins but generally suffer from a disrupted coding sequence
and an active homologous gene can be found at another locus. Sometimes these
entries have an intact coding sequence or an open but truncated ORF, in which
case there is other evidence used (for example genomic polyA stretches at
the 3' end) to classify them as a pseudogene. Can be further classified as one
of the following.

processed_pseudogene
----------------
Pseudogene that lack introns and is thought to arise from reverse transcription
of mRNA followed by reinsertion of DNA into the genome.

polymorphic_pseudogene
----------------
Pseudogene owing to a SNP/DIP but in other individuals/haplotypes/strains the
gene is translated.

retrotransposed
----------------
Pseudogene owing to a reverse transcribed and re-inserted sequence.

transcribed_processed_pseudogene
transcribed_unprocessed_pseudogene
transcribed_unitary_pseudogene
----------------
Pseudogene where protein homology or genomic structure indicates a pseudogene,
but the presence of locus-specific transcripts indicates expression.

translated_unprocessed_pseudogene
----------------
Pseudogene that has mass spec data suggesting that it is also translated.

unitary_pseudogene
----------------
A species specific unprocessed pseudogene without a parent gene, as it has an
active orthologue in another species.

unprocessed_pseudogene
----------------
Pseudogene that can contain introns since produced by gene duplication.

artifact
----------------
Used to tag mistakes in the public databases (Ensembl/SwissProt/Trembl)

lincRNA
----------------
Long, intervening noncoding (linc) RNAs, that can be found in evolutionarily
conserved, intergenic regions.

LRG_gene
----------------
Gene in a "Locus Reference Genomic" region known to have disease-related
sequence variations.

3prime_overlapping_ncrna
----------------
Transcripts where ditag and/or published experimental data strongly supports the
existence of short non-coding transcripts transcribed from the 3'UTR.

disrupted_domain
----------------
Otherwise viable coding region omitted from this alternatively spliced
transcript because the splice variation affects a region coding for a
protein domain.
"""

TCR_biotypes = {
    'TR_C_gene',
    'TR_D_gene',
    'TR_gene',
    'TR_J_gene',
    'TR_V_gene'
}

IG_biotypes = {
    'IG_C_gene',
    'IG_D_gene',
    'IG_gene',
    'IG_J_gene',
    'IG_LV_gene',
    'IG_M_gene',
    'IG_V_gene',
    'IG_Z_gene',
}

non_immune_protein_coding = {
    # premature stop codon will cause degradation
    'nonsense_mediated_decay',
    # TODO: find out what this is
    'nontranslating_CDS',
    # no stop codon within transcript, should get degraded at translation
    'non_stop_decay',
    # ordinary protein coding transcript
    'protein_coding',
    # usually a non-coding pseudogene but can be translated in some individuals
    # depending on common genetic variation
    'polymorphic_pseudogene',
    # Otherwise viable coding region omitted from this alternatively spliced
    # transcript because the splice variation affects a region coding for a
    # protein domain.
    'disrupted_domain',
    # Gene in a "Locus Reference Genomic" region known to have disease-related
    # sequence variations.
    'LRG_gene',
}

protein_coding = set.union(
    TCR_biotypes,
    IG_biotypes,
    non_immune_protein_coding
)


coding_pseudogenes = {
    # TODO: why is disrupted_domain a pseudogene rather than
    # a translated protein we expect to be dysfunctional?
    'disrupted_domain',
    'IG_C_pseudogene',
    'IG_J_pseudogene',
    'IG_pseudogene',
    'IG_V_pseudogene',
    # Found in ftp://ftp.ensembl.org/pub/release-81/gtf/mus_musculus
    'IG_D_pseudogene',
    'processed_pseudogene',
    'pseudogene',
    'transcribed_processed_pseudogene',
    'transcribed_unitary_pseudogene',
    'transcribed_unprocessed_pseudogene',
    'translated_processed_pseudogene',
    'TR_J_pseudogene',
    'TR_pseudogene',
    'TR_V_pseudogene',
    'unitary_pseudogene',
    'unprocessed_pseudogene',
    # to be experimentally confirmed
    'TEC',
    # TODO: should this be here or considered protein_coding?
    'translated_unprocessed_pseudogene',
    # pseudogene owing to a reverse transcribed and re-inserted sequence.
    "retrotransposed",
}

long_noncoding = {
    'lincRNA',
    'ncrna_host',
    '3prime_overlapping_ncrna',
    # why is an ambiguous ORF noncoding? Isn't rather coding but we can't
    # yet determine the sequence?
    'ambiguous_orf',
    'antisense',
    'antisense_RNA',
    'non_coding',
    'processed_transcript',
    'retained_intron',
    'sense_intronic',
    'sense_overlapping',
    'known_ncrna',
    # unspliced lncRNAs that are several kb in size.
    'macro_lncRNA',
}

mitochondrial = {
    'Mt_rRNA',
    'Mt_tRNA',
    'Mt_tRNA_pseudogene',
}

# short RNAs homologous to functional but which themselves are expected to
# be inert
short_noncoding_pseudogene = {
    'miRNA_pseudogene',
    'misc_RNA_pseudogene',
    'ncRNA_pseudogene',
    'rRNA_pseudogene',
    'scRNA_pseudogene',
    'snoRNA_pseudogene',
    'snRNA_pseudogene',
    'tRNA_pseudogene'
}

# short RNAs expected to have some effect or function in the nucleus
short_noncoding_functional = {
    'miRNA',
    'misc_RNA',
    'ncRNA',
    'rRNA',
    'scRNA',
    'snlRNA',
    'snoRNA',
    'snRNA',
    'tRNA',
    'sRNA',
    # Small Cajal body-specific RNA
    'scaRNA',
    # Vault RNA (http://en.wikipedia.org/wiki/Vault_RNA)
    'vaultRNA',
    'ribozyme',
}

short_noncoding = set.union(
    short_noncoding_functional,
    short_noncoding_pseudogene,
    mitochondrial)

valid_biotypes = set.union(
    protein_coding,
    coding_pseudogenes,
    long_noncoding,
    short_noncoding,
    # used to tag mistakes in the annotation database
    {"artifact"},
)

def is_valid_biotype(biotype):
    return biotype in valid_biotypes

def is_coding_biotype(biotype):
    return biotype in protein_coding



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from memoized_property import memoized_property

from .biotypes import is_valid_biotype
from .common import memoize
from .locus import Locus


class Transcript(Locus):
    """
    Transcript encompasses the locus, exons, and sequence of a transcript.

    Lazily fetches sequence in case we"re constructing many Transcripts
    and not using the sequence, avoid the memory/performance overhead
    of fetching and storing sequences from a FASTA file.
    """
    def __init__(
            self,
            transcript_id,
            transcript_name,
            contig,
            start,
            end,
            strand,
            biotype,
            gene_id,
            genome,
            require_valid_biotype=True):
        if require_valid_biotype and not is_valid_biotype(biotype):
            raise ValueError(
                "Invalid biotype '%s' for transcript with ID=%s, name=%s" % (
                    biotype, transcript_id, transcript_name))

        Locus.__init__(self, contig, start, end, strand)
        self.id = transcript_id
        self.name = transcript_name
        self.genome = genome
        self.db = genome.db
        self.biotype = biotype
        self.gene_id = gene_id

    def __str__(self):
        return (
            "Transcript(id=%s,"
            " name=%s,"
            " gene_id=%s,"
            " gene_name=%s,"
            " biotype=%s,"
            " location=%s:%d-%d)") % (
                    self.id,
                    self.name,
                    self.gene.id,
                    self.gene.name,
                    self.biotype,
                    self.contig,
                    self.start,
                    self.end)

    def __repr__(self):
        return str(self)

    def __len__(self):
        """
        Length of a transcript is the sum of its exon lengths
        """
        return sum(len(exon) for exon in self.exons)

    def __eq__(self, other):
        return (
            other.__class__ is Transcript and
            self.id == other.id and
            self.genome == other.genome)

    def __hash__(self):
        return hash(self.id)

    @memoized_property
    def gene(self):
        return self.genome.gene_by_id(self.gene_id)

    @memoized_property
    def exons(self):
        # need to look up exon_number alongside ID since each exon may
        # appear in multiple transcripts and have a different exon number
        # in each transcript
        columns = ["exon_number", "exon_id"]
        exon_numbers_and_ids = self.db.query(
            columns,
            filter_column="transcript_id",
            filter_value=self.id,
            feature="exon")

        # fill this list in its correct order (by exon_number) by using
        # the exon_number as a 1-based list offset
        exons = [None] * len(exon_numbers_and_ids)

        for exon_number, exon_id in exon_numbers_and_ids:
            exon = self.genome.exon_by_id(exon_id)
            exon_number = int(exon_number)
            assert exon_number >= 1, "Invalid exon number: %s" % exon_number
            assert exon_number <= len(exons), \
                "Invalid exon number: %s (max expected = %d)" % (
                    exon_number, len(exons))

            # exon_number is 1-based, convert to list index by subtracting 1
            exons[exon_number - 1] = exon

        assert all(exon is not None for exon in exons), \
            "Missing exons %s for transcript %s" % (
                [
                    i
                    for i, maybe_exon
                    in enumerate(exons) if maybe_exon is None
                ],
                self.id)
        return exons

    # possible annotations associated with transcripts
    _TRANSCRIPT_FEATURES = {"start_codon", "stop_codon", "UTR", "CDS"}

    @memoize
    def _transcript_feature_position_ranges(self, feature, required=True):
        """
        Find start/end chromosomal position range of features
        (such as start codon) for this transcript.
        """
        if feature not in self._TRANSCRIPT_FEATURES:
            raise ValueError("Invalid transcript feature: %s" % feature)

        results = self.db.query(
            select_column_names=["start", "end"],
            filter_column="transcript_id",
            filter_value=self.id,
            feature=feature)

        if required and len(results) == 0:
            raise ValueError(
                "Transcript %s does not contain feature %s" % (
                    self.id, feature))
        return results

    @memoize
    def _transcript_feature_positions(self, feature):
        """
        Get unique positions for feature, raise an error if feature is absent.
        """
        ranges = self._transcript_feature_position_ranges(
            feature, required=True)
        results = []
        # a feature (such as a stop codon), maybe be split over multiple
        # contiguous ranges. Collect all the nucleotide positions into a
        # single list.
        for (start, end) in ranges:
            # since ranges are [inclusive, inclusive] and
            # Python ranges are [inclusive, exclusive) we have to increment
            # the end position
            for position in range(start, end + 1):
                assert position not in results, \
                    "Repeated position %d for %s" % (position, feature)
                results.append(position)
        return results

    @memoize
    def _codon_positions(self, feature):
        """
        Parameters
        ----------
        feature : str
            Possible values are "start_codon" or "stop_codon"

        Returns list of three chromosomal positions.
        """
        results = self._transcript_feature_positions(feature)
        if len(results) != 3:
            raise ValueError(
                "Expected 3 positions for %s of %s but got %d" % (
                    feature,
                    self.id,
                    len(results)))
        return results

    @memoized_property
    def contains_start_codon(self):
        """
        Does this transcript have an annotated start_codon entry?
        """
        start_codons = self._transcript_feature_position_ranges(
            "start_codon", required=False)
        return len(start_codons) > 0

    @memoized_property
    def contains_stop_codon(self):
        """
        Does this transcript have an annotated stop_codon entry?
        """
        stop_codons = self._transcript_feature_position_ranges(
            "stop_codon", required=False)
        return len(stop_codons) > 0

    @memoized_property
    def start_codon_positions(self):
        """
        Chromosomal positions of nucleotides in start codon.
        """
        return self._codon_positions("start_codon")

    @memoized_property
    def stop_codon_positions(self):
        """
        Chromosomal positions of nucleotides in stop codon.
        """
        return self._codon_positions("stop_codon")

    @memoized_property
    def exon_intervals(self):
        """List of (start,end) tuples for each exon of this transcript,
        in the order specified by the 'exon_number' column of the
        exon table.
        """
        results = self.db.query(
            select_column_names=["exon_number", "start", "end"],
            filter_column="transcript_id",
            filter_value=self.id,
            feature="exon")
        sorted_intervals = [None] * len(results)
        for (exon_number, start, end) in results:
            sorted_intervals[int(exon_number) - 1] = (start, end)
        return sorted_intervals

    def spliced_offset(self, position):
        """
        Convert from an absolute chromosomal position to the offset into
        this transcript"s spliced mRNA.

        Position must be inside some exon (otherwise raise exception).
        """
        # this code is performance sensitive, so switching from
        # typechecks.require_integer to a simpler assertion
        assert type(position) == int, \
            "Position argument must be an integer, got %s : %s" % (
                position, type(position))

        if position < self.start or position > self.end:
            raise ValueError(
                "Invalid position: %d (must be between %d and %d)" % (
                    position,
                    self.start,
                    self.end))

        # offset from beginning of unspliced transcript (including introns)
        unspliced_offset = self.offset(position)
        total_spliced_offset = 0

        # traverse exons in order of their appearance on the strand
        # Since absolute positions may decrease if on the negative strand,
        # we instead use unspliced offsets to get always increasing indices.
        #
        # Example:
        #
        # Exon Name:                exon 1                exon 2
        # Spliced Offset:           123456                789...
        # Intron vs. Exon: ...iiiiiieeeeeeiiiiiiiiiiiiiiiieeeeeeiiiiiiiiiii...
        for exon in self.exons:
            exon_unspliced_start, exon_unspliced_end = self.offset_range(
                exon.start, exon.end)
            # If the relative position is not within this exon, keep a running
            # total of the total exonic length-so-far.
            #
            # Otherwise, if the relative position is within an exon, get its
            # offset into that exon by subtracting the exon"s relative start
            # position from the relative position. Add that to the total exonic
            # length-so-far.
            if exon_unspliced_start <= unspliced_offset <= exon_unspliced_end:
                # all offsets are base 0, can be used as indices into
                # sequence string
                exon_offset = unspliced_offset - exon_unspliced_start
                return total_spliced_offset + exon_offset
            else:
                exon_length = len(exon)  # exon_end_position - exon_start_position + 1
                total_spliced_offset += exon_length
        raise ValueError(
            "Couldn't find position %d on any exon of %s" % (
                position, self.id))

    @memoized_property
    def start_codon_unspliced_offsets(self):
        """
        Offsets from start of unspliced pre-mRNA transcript
        of nucleotides in start codon.
        """
        return [
            self.offset(position)
            for position
            in self.start_codon_positions
        ]

    @memoized_property
    def stop_codon_unspliced_offsets(self):
        """
        Offsets from start of unspliced pre-mRNA transcript
        of nucleotides in stop codon.
        """
        return [
            self.offset(position)
            for position
            in self.stop_codon_positions
        ]

    def _contiguous_offsets(self, offsets):
        """
        Sorts the input list of integer offsets,
        ensures that values are contiguous.
        """
        offsets.sort()
        for i in range(len(offsets) - 1):
            assert offsets[i] + 1 == offsets[i + 1], \
                "Offsets not contiguous: %s" % (offsets,)
        return offsets

    @memoized_property
    def start_codon_spliced_offsets(self):
        """
        Offsets from start of spliced mRNA transcript
        of nucleotides in start codon.
        """
        offsets = [
            self.spliced_offset(position)
            for position
            in self.start_codon_positions
        ]
        return self._contiguous_offsets(offsets)

    @memoized_property
    def stop_codon_spliced_offsets(self):
        """
        Offsets from start of spliced mRNA transcript
        of nucleotides in stop codon.
        """
        offsets = [
            self.spliced_offset(position)
            for position
            in self.stop_codon_positions
        ]
        return self._contiguous_offsets(offsets)

    @memoized_property
    def coding_sequence_position_ranges(self):
        """
        Return absolute chromosome position ranges for CDS fragments
        of this transcript
        """
        return self._transcript_feature_position_ranges("CDS")

    @memoized_property
    def complete(self):
        """
        Consider a transcript complete if it has start and stop codons and
        a coding sequence whose length is divisible by 3
        """
        return (
            self.contains_start_codon and
            self.contains_stop_codon and
            self.coding_sequence is not None and
            len(self.coding_sequence) % 3 == 0
        )

    @memoized_property
    def sequence(self):
        """
        Spliced cDNA sequence of transcript
        (includes 5" UTR, coding sequence, and 3" UTR)
        """
        return self.genome.transcript_sequences.get(self.id)

    @memoized_property
    def first_start_codon_spliced_offset(self):
        """
        Offset of first nucleotide in start codon into the spliced mRNA
        (excluding introns)
        """
        start_offsets = self.start_codon_spliced_offsets
        return min(start_offsets)

    @memoized_property
    def last_stop_codon_spliced_offset(self):
        """
        Offset of last nucleotide in stop codon into the spliced mRNA
        (excluding introns)
        """
        stop_offsets = self.stop_codon_spliced_offsets
        return max(stop_offsets)

    @memoized_property
    def coding_sequence(self):
        """
        cDNA coding sequence (from start codon to stop codon, without
        any introns)
        """
        start = self.first_start_codon_spliced_offset
        end = self.last_stop_codon_spliced_offset
        # If start codon is the at nucleotide offsets [3,4,5] and
        # stop codon is at nucleotide offsets  [20,21,22]
        # then start = 3 and end = 22.
        #
        # Adding 1 to end since Python uses non-inclusive ends in slices/ranges.

        # pylint: disable=invalid-slice-index
        # TODO(tavi) Figure out pylint is not happy with this slice
        return self.sequence[start:end + 1]

    @memoized_property
    def five_prime_utr_sequence(self):
        """
        cDNA sequence of 5' UTR
        (untranslated region at the beginning of the transcript)
        """
        # pylint: disable=invalid-slice-index
        # TODO(tavi) Figure out pylint is not happy with this slice
        return self.sequence[:self.first_start_codon_spliced_offset]

    @memoized_property
    def three_prime_utr_sequence(self):
        """
        cDNA sequence of 3' UTR
        (untranslated region at the end of the transcript)
        """
        return self.sequence[self.last_stop_codon_spliced_offset + 1:]

    @memoized_property
    def protein_id(self):
        result_tuple = self.db.query_one(
            select_column_names=["protein_id"],
            filter_column="transcript_id",
            filter_value=self.id,
            feature="CDS",
            distinct=True,
            required=False)
        if result_tuple:
            return result_tuple[0]
        else:
            return None

    @memoized_property
    def protein_sequence(self):
        if self.protein_id:
            return self.genome.protein_sequences.get(self.protein_id)
        else:
            return None



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the Genome class, with its millions of accessors and wrappers
around an arbitrary genomic database.
"""

from __future__ import print_function, division, absolute_import
import logging
from os import remove
from os.path import exists

from .common import memoize
from .memory_cache import MemoryCache
from .download_cache import DownloadCache
from .database import Database
from .exon import Exon
from .gene import Gene
from .gtf import GTF
from .sequence_data import SequenceData
from .transcript import Transcript

class Genome(object):
    """
    Bundles together the genomic annotation and sequence data associated with
    a particular genomic database source (e.g. a single Ensembl release) and
    provides a wide variety of helper methods for accessing this data.
    """
    def __init__(
            self,
            reference_name,
            annotation_name,
            annotation_version=None,
            gtf_path_or_url=None,
            transcript_fasta_path_or_url=None,
            protein_fasta_path_or_url=None,
            decompress_on_download=False,
            copy_local_files_to_cache=False,
            require_ensembl_ids=True,
            cache_directory_path=None):
        """
        Parameters
        ----------
        reference_name : str
            Name of genome assembly which annotations in GTF are aligned against
            (and from which sequence data is drawn)

        annotation_name : str
            Name of annotation source (e.g. "Ensembl)

        annotation_version : int or str
            Version of annotation database (e.g. 75)

        gtf_path_or_url : str
            Path or URL of GTF file

        transcript_fasta_path_or_url : str
            Path or URL of FASTA file containing transcript sequences

        protein_fasta_path_or_url : str
            Path or URL of FASTA file containing protein sequences

        decompress_on_download : bool
            If remote file is compressed, decompress the local copy?

        copy_local_files_to_cache : bool
            If genome data file is local use it directly or copy to cache first?

        require_ensembl_ids : bool
            Check gene/transcript/exon IDs to make sure they start with "ENS"

        cache_directory_path : None
            Where to place downloaded and cached files for this genome,
            by default inferred from reference name, annotation name,
            annotation version, and global cache dir for pyensembl.
        """

        self.reference_name = reference_name
        self.annotation_name = annotation_name
        self.annotation_version = annotation_version

        self.decompress_on_download = decompress_on_download
        self.copy_local_files_to_cache = copy_local_files_to_cache

        self.require_ensembl_ids = require_ensembl_ids

        self.download_cache = DownloadCache(
            reference_name=self.reference_name,
            annotation_name=self.annotation_name,
            annotation_version=self.annotation_version,
            decompress_on_download=self.decompress_on_download,
            copy_local_files_to_cache=self.copy_local_files_to_cache,
            install_string_function=self.install_string,
            cache_directory_path=cache_directory_path)
        self.cache_directory_path = self.download_cache.cache_directory_path

        self._gtf_path_or_url = gtf_path_or_url
        self.has_gtf = gtf_path_or_url is not None

        self._transcript_fasta_path_or_url = transcript_fasta_path_or_url
        self.has_transcript_fasta = transcript_fasta_path_or_url is not None

        self._protein_fasta_path_or_url = protein_fasta_path_or_url
        self.has_protein_fasta = protein_fasta_path_or_url is not None

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)
        self.memory_cache = MemoryCache()
        self._init_lazy_fields()

    def _init_lazy_fields(self):
        """
        Member data that gets loaded or constructed on demand
        """
        self._gtf = self._db = self.gtf_path = None
        self._protein_sequences = self.protein_fasta_path = None
        self._transcript_sequences = self.transcript_fasta_path = None

    def _get_cached_path(
            self,
            field_name,
            path_or_url,
            download_if_missing=False,
            overwrite=False):
        """
        Get the local path for a possibly remote file, invoking either
        a download or install error message if it's missing.
        """
        assert field_name, "Expected non-empty field name"
        assert path_or_url, "Expected non-empty path_or_url"
        return self.download_cache.local_path_or_install_error(
            field_name=field_name,
            path_or_url=path_or_url,
            download_if_missing=download_if_missing,
            overwrite=overwrite)

    def _get_gtf_path(self, download_if_missing=False, overwrite=False):
        return self._get_cached_path(
                field_name="gtf",
                path_or_url=self._gtf_path_or_url,
                download_if_missing=download_if_missing,
                overwrite=overwrite)

    def _get_transcript_fasta_path(
            self,
            download_if_missing=False,
            overwrite=False):
        if not self.has_transcript_fasta:
            raise ValueError("No transcript FASTA source for %s" % self)
        return self._get_cached_path(
            field_name="transcript-fasta",
            path_or_url=self._transcript_fasta_path_or_url,
            download_if_missing=download_if_missing,
            overwrite=overwrite)

    def _get_protein_fasta_path(
            self,
            download_if_missing=False,
            overwrite=False):
        # get the path for peptide FASTA files containing
        # this genome's protein sequences
        if not self.has_protein_fasta:
            raise ValueError("No protein FASTA source for %s" % self)
        return self._get_cached_path(
            field_name="protein-fasta",
            path_or_url=self._protein_fasta_path_or_url,
            download_if_missing=download_if_missing,
            overwrite=overwrite)

    def _set_local_paths(self, download_if_missing=False, overwrite=False):
        if self.has_gtf:
            self.gtf_path = self._get_gtf_path(
                download_if_missing=download_if_missing,
                overwrite=overwrite)
        if self.has_transcript_fasta:
            self.transcript_fasta_path = self._get_transcript_fasta_path(
                download_if_missing=download_if_missing,
                overwrite=overwrite)
        if self.has_protein_fasta:
            self.protein_fasta_path = self._get_protein_fasta_path(
                download_if_missing=download_if_missing,
                overwrite=overwrite)

    def download(self, overwrite=False):
        """
        Download data files needed by this Genome instance.

        Parameters
        ----------
        overwrite : bool, optional
            Download files regardless whether local copy already exists.
        """
        self._set_local_paths(download_if_missing=True, overwrite=overwrite)

    def index(self, overwrite=False):
        """
        Assuming that all necessary data for this Genome has been downloaded,
        generate the GTF database and save efficient representation of
        FASTA sequence files.
        """
        if self.has_gtf:
            self.db.connect_or_create(overwrite=overwrite)
        if self.has_transcript_fasta:
            self.transcript_sequences.index(overwrite=overwrite)
        if self.has_protein_fasta:
            self.protein_sequences.index(overwrite=overwrite)

    @property
    def gtf(self):
        if self._gtf is None:
            if not self.has_gtf:
                raise ValueError("Missing GTF source for %s" % self)
            # make sure GTF  file exists locally
            # and populate self.gtf_path
            self._set_local_paths()
            assert self.gtf_path is not None
            # GTF object wraps the source GTF file from which we get
            # genome annotations. Presents access to each feature
            # annotations as a pandas.DataFrame.
            self._gtf = GTF(
                gtf_path=self.gtf_path,
                cache_directory_path=self.cache_directory_path)
        return self._gtf

    @property
    def db(self):
        if self._db is None:
            # Database object turns the GTF dataframes into sqlite3 tables
            # and wraps them with methods like `query_one`
            self._db = Database(
                gtf=self.gtf,
                install_string=self.install_string())
        return self._db

    @property
    def protein_sequences(self):
        if self._protein_sequences is None:
            if not self.has_protein_fasta:
                raise ValueError(
                    "Missing protein FASTA source for %s" % self)
            # make sure protein FASTA file exists locally
            # and populate self.protein_fasta_path
            self._set_local_paths()
            assert self.protein_fasta_path is not None
            self._protein_sequences = SequenceData(
                fasta_path=self.protein_fasta_path,
                require_ensembl_ids=self.require_ensembl_ids,
                cache_directory_path=self.cache_directory_path)
        return self._protein_sequences

    @property
    def transcript_sequences(self):
        if self._transcript_sequences is None:
            if not self.has_transcript_fasta:
                raise ValueError(
                    "Missing transcript FASTA source for %s" % self)
            # make sure transcript FASTA file exists locally
            # and populate self.protein_fasta_path
            self._set_local_paths()
            assert self.transcript_fasta_path is not None
            self._transcript_sequences = SequenceData(
                fasta_path=self.transcript_fasta_path,
                require_ensembl_ids=self.require_ensembl_ids,
                cache_directory_path=self.cache_directory_path)
        return self._transcript_sequences

    def install_string(self):
        """
        Add every missing file to the install string shown to the user
        in an error message.
        """
        args = [
            "--reference-name", self.reference_name,
            "--annotation-name", self.annotation_name]
        if self.annotation_version:
            args.extend(["--annotation-version", str(self.annotation_version)])
        if self.has_gtf:
            args.append("--gtf")
            args.append("\"%s\"" % self._gtf_path_or_url)
        if self.has_protein_fasta:
            args.append("--protein-fasta")
            args.append("\"%s\"" % self._protein_fasta_path_or_url)
        if self.has_transcript_fasta:
            args.append("--transcript-fasta")
            args.append("\"%s\"" % self._transcript_fasta_path_or_url)
        return "pyensembl install %s" % " ".join(args)

    def __str__(self):
        return ("Genome(reference_name=%s, "
                "annotation_name=%s, "
                "annotation_version=%s, "
                "gtf_path_or_url=%s, "
                "transcript_fasta_path_or_url=%s, "
                "protein_fasta_path_or_url=%s)" % (
                    self.reference_name,
                    self.annotation_name,
                    self.annotation_version,
                    self._gtf_path_or_url,
                    self._transcript_fasta_path_or_url,
                    self._protein_fasta_path_or_url))

    def __repr__(self):
        return str(self)

    def _fields(self):
        return (
            self.reference_name,
            self.annotation_name,
            self.annotation_version,
            self._gtf_path_or_url,
            self._protein_fasta_path_or_url,
            self._transcript_fasta_path_or_url,
        )

    def __eq__(self, other):
        return (
            other.__class__ is Genome and
            self._fields() == other._fields()
        )

    def __hash__(self):
        return hash(self._fields())

    def clear_cache(self):
        """
        Clear any in-memory cached values and short-lived on-disk
        materializations from MemoryCache
        """
        for maybe_fn in self.__dict__.values():
            # clear cache associated with all memoization decorators,
            # GTF and SequenceData objects
            if hasattr(maybe_fn, "clear_cache"):
                maybe_fn.clear_cache()

    def delete_index_files(self):
        """
        Delete all data aside from source GTF and FASTA files
        """
        self.clear_cache()
        db_path = self.db.local_db_path()
        if exists(db_path):
            remove(db_path)

    def all_feature_values(
            self,
            column,
            feature,
            distinct=True,
            contig=None,
            strand=None):
        """
        Cached lookup of all values for a particular feature property from
        the database, caches repeated queries in memory and
        stores them as a CSV.

        Parameters
        ----------

        column : str
            Name of property (e.g. exon_id)

        feature : str
            Type of entry (e.g. exon)

        distinct : bool, optional
            Keep only unique values

        contig : str, optional
            Restrict query to particular contig

        strand : str, optional
            Restrict results to "+" or "-" strands

        Returns a list constructed from query results.
        """
        if not self.gtf:
            raise ValueError("No GTF supplied to this Genome: %s" %
                             str(self))

        # since we're constructing a list, rather than a DataFrame,
        # we're going to store it using pickling rather than the Pandas
        # CSV serializer. Change the default extension from ".csv" to ".pickle"
        pickle_path = self.gtf.data_subset_path(
            feature=feature,
            column=column,
            contig=contig,
            strand=strand,
            distinct=distinct,
            extension=".pickle")

        def run_query():
            results = self.db.query_feature_values(
                column=column,
                feature=feature,
                distinct=distinct,
                contig=contig,
                strand=strand)
            assert isinstance(results, list), \
                "Expected list from Database.query_feature_values, got %s" % (
                    type(results))
            return results

        return self.memory_cache.cached_object(
            pickle_path, compute_fn=run_query)

    def transcript_sequence(self, transcript_id):
        """Return cDNA nucleotide sequence of transcript, or None if
        transcript doesn't have cDNA sequence.
        """
        if self.transcript_sequences is None:
            raise ValueError(
                "No transcript FASTA supplied to this Genome: %s" % self)
        return self.transcript_sequences.get(transcript_id)

    def protein_sequence(self, protein_id):
        """Return cDNA nucleotide sequence of transcript, or None if
        transcript doesn't have cDNA sequence.
        """
        if self.protein_sequences is None:
            raise ValueError(
                "No protein FASTA supplied to this Genome: %s" % self)
        return self.protein_sequences.get(protein_id)

    def genes_at_locus(self, contig, position, end=None, strand=None):
        gene_ids = self.gene_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    def transcripts_at_locus(self, contig, position, end=None, strand=None):
        transcript_ids = self.transcript_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    def exons_at_locus(self, contig, position, end=None, strand=None):
        exon_ids = self.exon_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [
            self.exon_by_id(exon_id)
            for exon_id in exon_ids
        ]

    def gene_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="gene_id",
            feature="gene",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def gene_names_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
             column="gene_name",
             feature="gene",
             contig=contig,
             position=position,
             end=end,
             strand=strand)

    def exon_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="exon_id",
            feature="exon",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def transcript_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="transcript_id",
            feature="transcript",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def transcript_names_at_locus(
            self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="transcript_name",
            feature="transcript",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def protein_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="protein_id",
            feature="transcript",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    ###################################################
    #
    #         Methods which return Locus objects
    #         containing (contig, start, stop, strand)
    #         of various genomic entities
    #
    ###################################################

    @memoize
    def locus_of_gene_id(self, gene_id):
        """
        Given a gene ID returns Locus with: chromosome, start, stop, strand
        """
        return self.db.query_locus(
            filter_column="gene_id",
            filter_value=gene_id,
            feature="gene")

    @memoize
    def loci_of_gene_names(self, gene_name):
        """
        Given a gene name returns list of Locus objects with fields:
            chromosome, start, stop, strand
        You can get multiple results since a gene might have multiple copies
        in the genome.
        """
        return self.db.query_loci("gene_name", gene_name, "gene")

    @memoize
    def locus_of_transcript_id(self, transcript_id):
        return self.db.query_locus(
            filter_column="transcript_id",
            filter_value=transcript_id,
            feature="transcript")

    @memoize
    def locus_of_exon_id(self, exon_id):
        """
        Given an exon ID returns Locus
        """
        return self.db.query_locus("exon_id", exon_id, feature="exon")

    ###################################################
    #
    #             Gene Info Objects
    #
    ###################################################

    @memoize
    def genes(self, contig=None, strand=None):
        """
        Returns all Gene objects in the database. Can be restricted to a
        particular contig/chromosome and strand by the following arguments:

        Parameters
        ----------
        contig : str
            Only return genes on the given contig.

        strand : str
            Only return genes on this strand.
        """
        gene_ids = self.gene_ids(contig=contig, strand=strand)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    @memoize
    def gene_by_id(self, gene_id):
        """
        Construct a Gene object for the given gene ID.
        """
        field_names = [
            "seqname",
            "start",
            "end",
            "strand",
        ]
        optional_field_names = [
            "gene_name",
            "gene_biotype",
        ]
        # Do not look for gene_name and gene_biotype if they are
        # not in the database.
        field_names.extend([name for name in optional_field_names
                            if self.db.column_exists("gene", name)])
        result = self.db.query_one(
            field_names,
            filter_column="gene_id",
            filter_value=gene_id,
            feature="gene")
        if not result:
            raise ValueError("Gene not found: %s" % (gene_id,))

        gene_name, gene_biotype = None, None
        assert len(result) >= 4 and len(result) <= 6, \
            "Result is not the expected length: %d" % len(result)
        contig, start, end, strand = result[:4]
        if len(result) == 5:
            if "gene_name" in field_names:
                gene_name = result[4]
            else:
                gene_biotype = result[4]
        elif len(result) == 6:
            gene_name, gene_biotype = result[4:]

        return Gene(
            gene_id=gene_id,
            gene_name=gene_name,
            contig=contig,
            start=start,
            end=end,
            strand=strand,
            biotype=gene_biotype,
            genome=self,
            require_valid_biotype=("gene_biotype" in field_names))

    @memoize
    def genes_by_name(self, gene_name):
        """
        Get all the unqiue genes with the given name (there might be multiple
        due to copies in the genome), return a list containing a Gene object
        for each distinct ID.
        """
        gene_ids = self.gene_ids_of_gene_name(gene_name)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    @memoize
    def gene_by_protein_id(self, protein_id):
        """
        Get the gene ID associated with the given protein ID,
        return its Gene object
        """
        gene_id = self.gene_id_of_protein_id(protein_id)
        return self.gene_by_id(gene_id)

    ###################################################
    #
    #             Gene Names
    #
    ###################################################

    @memoize
    def _query_gene_name(self, property_name, property_value, feature_type):
        results = self.db.query(
            select_column_names=["gene_name"],
            filter_column=property_name,
            filter_value=property_value,
            feature=feature_type,
            distinct=True,
            required=True)
        return str(results[0][0])

    @memoize
    def gene_names(self, contig=None, strand=None):
        """
        Return all genes in the database,
        optionally restrict to a chromosome and/or strand.
        """
        return self.all_feature_values(
            column="gene_name",
            feature="gene",
            contig=contig,
            strand=strand)

    @memoize
    def gene_name_of_gene_id(self, gene_id):
        return self._query_gene_name("gene_id", gene_id, "gene")

    @memoize
    def gene_name_of_transcript_id(self, transcript_id):
        return self._query_gene_name(
            "transcript_id", transcript_id, "transcript")

    @memoize
    def gene_name_of_transcript_name(self, transcript_name):
        return self._query_gene_name(
            "transcript_name", transcript_name, "transcript")

    @memoize
    def gene_name_of_exon_id(self, exon_id):
        return self._query_gene_name("exon_id", exon_id, "exon")

    ###################################################
    #
    #             Gene IDs
    #
    ###################################################

    @memoize
    def _query_gene_ids(self, property_name, value, feature="gene"):
        results = self.db.query(
            select_column_names=["gene_id"],
            filter_column=property_name,
            filter_value=value,
            feature=feature,
            distinct=True,
            required=True)
        return [str(result[0]) for result in results if result[0]]

    @memoize
    def gene_ids(self, contig=None, strand=None):
        """
        What are all the gene IDs
        (optionally restrict to a given chromosome/contig and/or strand)
        """
        return self.all_feature_values(
            column="gene_id",
            feature="gene",
            contig=contig,
            strand=strand)

    @memoize
    def gene_ids_of_gene_name(self, gene_name):
        """
        What are the gene IDs associated with a given gene name?
        (due to copy events, there might be multiple genes per name)
        """
        results = self._query_gene_ids("gene_name", gene_name)
        if len(results) == 0:
            raise ValueError("Gene name not found: %s" % gene_name)
        return results

    @memoize
    def gene_id_of_protein_id(self, protein_id):
        """
        What is the gene ID associated with a given protein ID?
        """
        results = self._query_gene_ids("protein_id", protein_id,
                                       feature="CDS")
        if len(results) == 0:
            raise ValueError("Protein ID not found: %s" % protein_id)
        assert len(results) == 1, \
            ("Should have only one gene ID for a given protein ID, "
             "but found %d: %s" % (len(results), results))
        return results[0]

    ###################################################
    #
    #             Transcript Info Objects
    #
    ###################################################

    @memoize
    def transcripts(self, contig=None, strand=None):
        """
        Construct Transcript object for every transcript entry in
        the database. Optionally restrict to a particular
        chromosome using the `contig` argument.
        """
        transcript_ids = self.transcript_ids(contig=contig, strand=strand)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    @memoize
    def transcript_by_id(self, transcript_id):
        """Construct Transcript object with given transcript ID"""

        optional_field_names = [
            "transcript_name",
            "transcript_biotype",
        ]
        field_names = [
            "seqname",
            "start",
            "end",
            "strand",
            "gene_id",
        ]
        # Do not look for transcript_name and transcript_biotype if
        # they are not in the database.
        field_names.extend([name for name in optional_field_names
                            if self.db.column_exists("transcript", name)])
        result = self.db.query_one(
            select_column_names=field_names,
            filter_column="transcript_id",
            filter_value=transcript_id,
            feature="transcript",
            distinct=True)
        if not result:
            raise ValueError("Transcript not found: %s" % (transcript_id,))

        transcript_name, transcript_biotype = None, None
        assert len(result) >= 5 and len(result) <= 7, \
            "Result is not the expected length: %d" % len(result)
        contig, start, end, strand, gene_id = result[:5]
        if len(result) == 6:
            if "transcript_name" in field_names:
                transcript_name = result[5]
            else:
                transcript_biotype = result[5]
        elif len(result) == 7:
            transcript_name, transcript_biotype = result[5:]

        return Transcript(
            transcript_id=transcript_id,
            transcript_name=transcript_name,
            contig=contig,
            start=start,
            end=end,
            strand=strand,
            biotype=transcript_biotype,
            gene_id=gene_id,
            genome=self,
            require_valid_biotype=("transcript_biotype" in field_names))

    @memoize
    def transcripts_by_name(self, transcript_name):
        transcript_ids = self.transcript_ids_of_transcript_name(transcript_name)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    @memoize
    def transcript_by_protein_id(self, protein_id):
        transcript_id = self.transcript_id_of_protein_id(protein_id)
        return self.transcript_by_id(transcript_id)

    ###################################################
    #
    #            Transcript Names
    #
    ###################################################

    @memoize
    def _query_transcript_names(self, property_name, value):
        results = self.db.query(
            select_column_names=["transcript_name"],
            filter_column=property_name,
            filter_value=value,
            feature="transcript",
            distinct=True,
            required=True)
        return [result[0] for result in results]

    @memoize
    def transcript_names(self, contig=None, strand=None):
        """
        What are all the transcript names in the database
        (optionally, restrict to a given chromosome and/or strand)
        """
        return self.all_feature_values(
            column="transcript_name",
            feature="transcript",
            contig=contig,
            strand=strand)

    @memoize
    def transcript_names_of_gene_name(self, gene_name):
        return self._query_transcript_names("gene_name", gene_name)

    @memoize
    def transcript_name_of_transcript_id(self, transcript_id):
        transcript_names = self._query_transcript_names(
            "transcript_id", transcript_id)
        if len(transcript_names) == 0:
            raise ValueError(
                "No transcript names for transcript ID = %s" % transcript_id)
        assert len(transcript_names) == 1, \
            "Multiple transcript names for transcript ID = %s" % transcript_id
        return transcript_names[0]

    ###################################################
    #
    #            Transcript IDs
    #
    ###################################################

    @memoize
    def _query_transcript_ids(
            self,
            property_name,
            value,
            feature="transcript"):
        results = self.db.query(
            select_column_names=["transcript_id"],
            filter_column=property_name,
            filter_value=value,
            feature=feature,
            distinct=True,
            required=True)
        return [result[0] for result in results]

    @memoize
    def transcript_ids(self, contig=None, strand=None):
        return self.all_feature_values(
            column="transcript_id",
            feature="transcript",
            contig=contig,
            strand=strand)

    @memoize
    def transcript_ids_of_gene_id(self, gene_id):
        return self._query_transcript_ids("gene_id", gene_id)

    @memoize
    def transcript_ids_of_gene_name(self, gene_name):
        return self._query_transcript_ids("gene_name", gene_name)

    @memoize
    def transcript_ids_of_transcript_name(self, transcript_name):
        return self._query_transcript_ids("transcript_name", transcript_name)

    @memoize
    def transcript_ids_of_exon_id(self, exon_id):
        return self._query_transcript_ids("exon_id", exon_id)

    @memoize
    def transcript_id_of_protein_id(self, protein_id):
        """
        What is the transcript ID associated with a given protein ID?
        """
        results = self._query_transcript_ids("protein_id", protein_id,
                                             feature="CDS")
        if len(results) == 0:
            raise ValueError("Protein ID not found: %s" % protein_id)
        assert len(results) == 1, \
            ("Should have only one transcript ID for a given protein ID, "
             "but found %d: %s" % (len(results), results))
        return results[0]

    ###################################################
    #
    #             Exon Info Objects
    #
    ###################################################

    @memoize
    def exons(self, contig=None, strand=None):
        """
        Create exon object for all exons in the database, optionally
        restrict to a particular chromosome using the `contig` argument.
        """
        # DataFrame with single column called "exon_id"
        exon_ids = self.exon_ids(contig=contig, strand=strand)
        return [
            self.exon_by_id(exon_id)
            for exon_id in exon_ids
        ]

    @memoize
    def exon_by_id(self, exon_id):
        """Construct an Exon object from its ID by looking up the exon"s
        properties in the given Database.
        """
        field_names = [
            "seqname",
            "start",
            "end",
            "strand",
            "gene_name",
            "gene_id",
        ]

        contig, start, end, strand, gene_name, gene_id = self.db.query_one(
            select_column_names=field_names,
            filter_column="exon_id",
            filter_value=exon_id,
            feature="exon",
            distinct=True)

        return Exon(
            exon_id=exon_id,
            contig=contig,
            start=start,
            end=end,
            strand=strand,
            gene_name=gene_name,
            gene_id=gene_id)

    ###################################################
    #
    #                Exon IDs
    #
    ###################################################

    @memoize
    def _query_exon_ids(self, property_name, value):
        results = self.db.query(
            select_column_names=["exon_id"],
            filter_column=property_name,
            filter_value=value,
            feature="exon",
            distinct=True,
            required=True)
        return [result[0] for result in results]

    @memoize
    def exon_ids(self, contig=None, strand=None):
        return self.all_feature_values(
            column="exon_id",
            feature="exon",
            contig=contig,
            strand=strand)

    @memoize
    def exon_ids_of_gene_id(self, gene_id):
        return self._query_exon_ids("gene_id", gene_id)

    @memoize
    def exon_ids_of_gene_name(self, gene_name):
        return self._query_exon_ids("gene_name", gene_name)

    @memoize
    def exon_ids_of_transcript_name(self, transcript_name):
        return self._query_exon_ids("transcript_name", transcript_name)

    @memoize
    def exon_ids_of_transcript_id(self, transcript_id):
        return self._query_exon_ids("transcript_id", transcript_id)

    ###################################################
    #
    #             Protein IDs
    #
    ###################################################

    @memoize
    def protein_ids(self, contig=None, strand=None):
        """
        What are all the protein IDs
        (optionally restrict to a given chromosome and/or strand)
        """
        protein_ids = self.all_feature_values(
            column="protein_id",
            feature="CDS",
            contig=contig,
            strand=strand,
            distinct=True)
        # drop None values
        return [protein_id for protein_id in protein_ids if protein_id]


# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

import logging
from os.path import join, exists
import sqlite3

import datacache
from typechecks import require_integer, require_string

from .common import memoize
from .locus import normalize_chromosome, normalize_strand, Locus

# any time we update the database schema, increment this version number
DATABASE_SCHEMA_VERSION = 2

class Database(object):
    """
    Wrapper around sqlite3 database so that the rest of the
    library doesn't have to worry about constructing the .db file or
    writing SQL queries directly.
    """

    def __init__(self, gtf, install_string):
        """
        Parameters
        ----------
        gtf : pyensembl.GTF instance
            Object which parses GTF annotation files and presents their
            contents as Pandas DataFrames

        install_string : str
            Message to tell user if database connection is requested before
            database is created.
        """
        self.gtf = gtf
        self.install_string = install_string
        self._connection = None

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)

    def __eq__(self, other):
        return (
            other.__class__ is Database and
            self.gtf == other.gtf)

    def __str__(self):
        return "Database(gtf=%s)" % (self.gtf,)

    def __hash__(self):
        return hash((self.gtf))

    def local_db_filename(self):
        if not self.gtf:
            raise ValueError("No GTF supplied to this Database: %s" %
                             str(self))
        base = self.gtf.gtf_base_filename
        return base + ".db"

    def local_db_path(self):
        if not self.gtf:
            raise ValueError("No GTF supplied to this Database: %s" %
                             str(self))
        dirpath = self.gtf.cache_directory_path
        filename = self.local_db_filename()
        return join(dirpath, filename)

    def _all_possible_indices(self, column_names):
        """
        Create list of tuples containing all possible index groups
        we might want to create over tables in this database.

        If a set of genome annotations is missing some column we want
        to index on, we have to drop any indices which use that column.

        A specific table may later drop some of these indices if they're
        missing  values for that feature or are the same as the table's primary key.
        """
        candidate_column_groups = [
            ['seqname', 'start', 'end'],
            ['seqname', 'start', 'end', 'strand'],
            ['gene_name'],
            ['gene_id'],
            ['transcript_id'],
            ['transcript_name'],
            ['exon_id'],
            ['protein_id'],
            ['ccds_id'],
        ]
        indices = []
        column_set = set(column_names)
        # Since queries are often restricted by feature type
        # we should include that column in combination with all
        # other indices we anticipate might improve performance
        for column_group in candidate_column_groups:
            skip = False
            for column_name in column_group:
                # some columns, such as 'exon_id',
                # are not available in all releases of Ensembl (or
                # other GTFs)
                if column_name not in column_set:
                    logging.info(
                        "Skipping database index for {%s}",
                        ", ".join(column_group))
                    skip = True
            if skip:
                continue
            indices.append(column_group)
        return indices

    # mapping from database tables to their primary keys
    # sadly exon IDs *are* not unique, so can't be in this dict
    PRIMARY_KEY_COLUMNS = {
        'gene': 'gene_id',
        'transcript': 'transcript_id'
    }

    def _get_primary_key(self, feature_name, feature_df):
        """Name of primary key for a feature table (e.g. "gene" -> "gene_id")

        Since we're potentially going to run this code over unseen data,
        make sure that the primary is unique and never null.

        If a feature doesn't have a primary key, return None.
        """
        if feature_name not in self.PRIMARY_KEY_COLUMNS:
            return None
        primary_key = self.PRIMARY_KEY_COLUMNS[feature_name]
        primary_key_values = feature_df[primary_key]
        if primary_key_values.isnull().any():
            raise ValueError(
                "Column '%s' can't be primary key of table '%s'"
                " because it contains nulls values" % (
                    primary_key, feature_name))
        elif len(primary_key_values.unique()) < len(primary_key_values):
            raise ValueError(
                "Column '%s' can't be primary key of table '%s'"
                " because it contains repeated values" % (
                    primary_key, feature_name))
        else:
            return primary_key

    def _feature_indices(self, all_index_groups, primary_key, feature_df):
        """Choose subset ofindex group tuples from `all_index_groups` which are
        applicable to a particular feature (not same as its primary key, have
        non-null values).
        """
        # each feature only gets indices if they're *not* the
        # primary key and have non-null values in the feature's
        # subset of data
        result = []
        for index_group in all_index_groups:
            # is the index group just a primary key?
            if len(index_group) == 1 and index_group[0] == primary_key:
                continue
            index_column_values = feature_df[index_group]
            if len(index_column_values.dropna()) == 0:
                continue
            result.append(index_group)
        return result

    def create(self, overwrite=False):
        """
        Create the local database (including indexing) if it's not
        already set up. If `overwrite` is True, always re-create
        the database from scratch.

        Returns a connection to the database.
        """
        if not self.gtf:
            raise ValueError("No GTF supplied to this Database: %s" %
                             str(self))

        db_path = self.local_db_path()
        print("Creating database: %s" % (db_path,))
        df = self.gtf.dataframe()
        all_index_groups = self._all_possible_indices(df.columns)

        # split single DataFrame into dictionary mapping each unique
        # feature name onto that subset of the data
        feature_names = df['feature'].unique()
        dataframes = {}
        # every table gets the same set of indices
        indices_dict = {}
        # if a feature has an ID then make it that table's primary key
        primary_keys = {}

        for feature in feature_names:
            df_subset = df[df.feature == feature]
            dataframes[feature] = df_subset

            primary_key = self._get_primary_key(feature, df_subset)
            if primary_key:
                primary_keys[feature] = primary_key

            indices_dict[feature] = self._feature_indices(
                all_index_groups,
                primary_key,
                df_subset)
        self._connection = datacache.db_from_dataframes_with_absolute_path(
            db_path=db_path,
            table_names_to_dataframes=dataframes,
            table_names_to_primary_keys=primary_keys,
            table_names_to_indices=indices_dict,
            overwrite=overwrite,
            version=DATABASE_SCHEMA_VERSION)
        return self._connection

    def _get_connection(self):
        if self._connection is None:
            db_path = self.local_db_path()
            if exists(db_path):
                # since version metadata is filled in last in datacache,
                # checking for a version also implicitly checks for
                # having finishing fill in all the tables/rows.
                #
                # TODO: expose this more explicitly in datacache
                #
                self._connection = datacache.connect_if_correct_version(
                    db_path, DATABASE_SCHEMA_VERSION)
        return self._connection

    @property
    def connection(self):
        """
        Get a connection to the database or raise an exception
        """
        connection = self._get_connection()
        if connection:
            return connection
        else:
            raise ValueError(
                "GTF database needs to be created, run: %s" % (
                    self.install_string,))

    def connect_or_create(self, overwrite=False):
        """
        Return a connection to the database if it exists, otherwise create it.
        Overwrite the existing database if `overwrite` is True.
        """
        connection = self._get_connection()
        if connection:
            return connection
        else:
            return self.create(overwrite=overwrite)

    @memoize
    def columns(self, table_name):
        sql = "PRAGMA table_info(%s)" % table_name
        table_info = self.connection.execute(sql).fetchall()
        return [info[1] for info in table_info]

    @memoize
    def column_exists(self, table_name, column_name):
        return column_name in self.columns(table_name)

    def column_values_at_locus(
            self,
            column_name,
            feature,
            contig,
            position,
            end=None,
            strand=None,
            distinct=False,
            sorted=False):
        """
        Get the non-null values of a column from the database
        at a particular range of loci
        """

        # TODO: combine with the query method, since they overlap
        # significantly
        require_string(column_name, "column_name", nonempty=True)

        contig = normalize_chromosome(contig)

        require_integer(position, "position")

        if end is None:
            end = position

        require_integer(end, "end")

        if not self.column_exists(feature, column_name):
            raise ValueError("Table %s doesn't have column %s" % (
                feature, column_name,))

        if distinct:
            distinct_string = "DISTINCT "
        else:
            distinct_string = ""

        query = """
            SELECT %s%s
            FROM %s
            WHERE seqname = ?
            AND start <= ?
            AND end >= ?

        """ % (distinct_string, column_name, feature)

        query_params = [contig, end, position]

        if strand:
            query += " AND strand = ?"
            query_params.append(strand)

        tuples = self.connection.execute(query, query_params).fetchall()

        # each result is a tuple, so pull out its first element
        results = [t[0] for t in tuples if t[0] is not None]

        if sorted:
            results.sort()
        return results

    def distinct_column_values_at_locus(
            self,
            column,
            feature,
            contig,
            position,
            end=None,
            strand=None):
        """
        Gather all the distinct values for a property/column at some specified
        locus.

        Parameters
        ----------
        column : str
            Which property are we getting the values of.

        feature : str
            Which type of entry (e.g. transcript, exon, gene) is the property
            associated with?

        contig : str
            Chromosome or unplaced contig name

        position : int
            Chromosomal position

        end : int, optional
            End position of a range, if unspecified assume we're only looking
            at the single given position.

        strand : str, optional
            Either the positive ('+') or negative strand ('-'). If unspecified
            then check for values on either strand.
        """
        return self.column_values_at_locus(
            column,
            feature,
            contig,
            position,
            end=end,
            strand=strand,
            distinct=True,
            sorted=True)

    def run_sql_query(self, sql, required=False, query_params=[]):
        """
        Given an arbitrary SQL query, run it against the database
        and return the results.

        Parameters
        ----------
        sql : str
            SQL query

        required : bool
            Raise an error if no results found in the database

        query_params : list
            For each '?' in the query there must be a corresponding value in
            this list.
        """
        try:
            cursor = self.connection.execute(sql, query_params)
        except sqlite3.OperationalError as e:
            logging.warn(
                "Encountered error \"%s\" from query \"%s\" with parameters %s",
                e.message, sql, query_params)
            raise
        results = cursor.fetchall()
        if required and not results:
            raise ValueError(
                "No results found for query:\n%s\nwith parameters: %s" % (
                    sql, query_params))

        return results

    @memoize
    def query(
            self,
            select_column_names,
            filter_column,
            filter_value,
            feature,
            distinct=False,
            required=False):
        """
        Construct a SQL query and run against the sqlite3 database,
        filtered both by the feature type and a user-provided column/value.
        """
        sql = """
            SELECT %s%s
            FROM %s
            WHERE %s = ?
        """ % ("distinct " if distinct else "",
               ", ".join(select_column_names),
               feature,
               filter_column)
        query_params = [filter_value]
        return self.run_sql_query(
            sql, required=required, query_params=query_params)

    @memoize
    def query_one(
            self,
            select_column_names,
            filter_column,
            filter_value,
            feature,
            distinct=False,
            required=False):
        results = self.query(
            select_column_names,
            filter_column,
            filter_value,
            feature,
            distinct=distinct,
            required=required)

        if len(results) == 0:
            if required:
                raise ValueError("%s not found: %s" % (
                    filter_column, filter_value))
            else:
                return None
        elif len(results) > 1:
            raise ValueError(
                "Found multiple entries with %s=%s (%s)" % (
                    filter_column, filter_value, results))
        return results[0]

    @memoize
    def query_feature_values(
            self,
            column,
            feature,
            distinct=True,
            contig=None,
            strand=None):
        """
        Run a SQL query against the sqlite3 database, filtered
        only on the feature type.
        """
        query = """
            SELECT %s%s
            FROM %s
            WHERE 1=1
        """ % ("DISTINCT " if distinct else "", column, feature)
        query_params = []

        if contig:
            contig = normalize_chromosome(contig)
            query += " AND seqname = ?"
            query_params.append(contig)

        if strand:
            strand = normalize_strand(strand)
            query += " AND strand = ?"
            query_params.append(strand)

        rows = self.run_sql_query(query, query_params=query_params)
        return [row[0] for row in rows if row is not None]

    @memoize
    def query_distinct_on_contig(self, column_name, feature, contig):
        return self.query_feature_values(
            column=column_name,
            feature=feature,
            contig=contig,
            distinct=True)

    @memoize
    def query_loci(self, filter_column, filter_value, feature):
        """
        Query for loci satisfying a given filter and feature type.


        Parameters
        ----------
        filter_column : str
            Name of column to filter results by.

        filter_value : str
            Only return loci which have this value in the their filter_column.

        feature : str
            Feature names such as 'transcript', 'gene', and 'exon'

        Returns list of Locus objects
        """
        # list of values containing (contig, start, stop, strand)
        result_tuples = self.query(
            select_column_names=["seqname", "start", "end", "strand"],
            filter_column=filter_column,
            filter_value=filter_value,
            feature=feature,
            distinct=True,
            required=True)
        return [
            Locus(contig, start, end, strand)
            for (contig, start, end, strand)
            in result_tuples
        ]

    @memoize
    def query_locus(self, filter_column, filter_value, feature):
        """
        Query for unique locus, raises error if missing or more than
        one locus in the database.

        Parameters
        ----------
        filter_column : str
            Name of column to filter results by.

        filter_value : str
            Only return loci which have this value in the their filter_column.

        feature : str
            Feature names such as 'transcript', 'gene', and 'exon'

        Returns single Locus object.
        """
        loci = self.query_loci(
            filter_column=filter_column,
            filter_value=filter_value,
            feature=feature)

        if len(loci) == 0:
            raise ValueError("Couldn't find locus for %s with %s = %s" % (
                feature, filter_column, filter_value))
        elif len(loci) > 1:
            raise ValueError("Too many loci for %s with %s = %s: %s" % (
                feature, filter_column, filter_value, loci))
        return loci[0]




# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, absolute_import, division

from .ensembl_release_versions import MAX_ENSEMBL_RELEASE


class Species(object):
    """
    Container for combined information about a species name, its synonyn names
    and which reference to use for this species in each Ensembl release.
    """
    def __init__(self, latin_name, synonyms=[], reference_assemblies={}):
        """
        Parameters
        ----------
        latin_name : str

        synonyms : list of strings

        reference_assemblies : dict
            Mapping of names of reference genomes onto inclusive ranges of
            Ensembl releases Example: {"GRCh37": (54, 75)}
        """
        self.latin_name = latin_name.lower().replace(" ", "_")
        self.synonyms = synonyms
        self.reference_assemblies = reference_assemblies
        self._release_to_genome = {}
        for (genome_name, (start, end)) in self.reference_assemblies.items():
            for i in range(start, end + 1):
                assert i not in self._release_to_genome, \
                    "Ensembl release %d already has an associated genome"
                self._release_to_genome[i] = genome_name

    def which_reference(self, ensembl_release):
        if ensembl_release not in self._release_to_genome:
            raise ValueError("No genome for %s in Ensembl release %d" % (
                self.latin_name, ensembl_release))
        return self._release_to_genome[ensembl_release]

    def __str__(self):
        return (
            "Species(latin_name='%s', synonyms=%s, reference_assemblies=%s)" % (
                self.latin_name, self.synonyms, self.reference_assemblies))

    def __repr__(self):
        return str(self)

_latin_names_to_species = {}
_common_names_to_species = {}
_reference_names_to_species = {}

def add_species(latin_name, synonyms, reference_assemblies):
    """
    Create a Species object from the given arguments and enter into
    all the dicts used to look the species up by its fields.
    """
    species = Species(
        latin_name=latin_name,
        synonyms=synonyms,
        reference_assemblies=reference_assemblies)
    _latin_names_to_species[species.latin_name] = species
    for synonym in synonyms:
        if synonym in _common_names_to_species:
            raise ValueError("Can't use synonym '%s' for both %s and %s" % (
                synonym,
                species,
                _common_names_to_species[synonym]))
        _common_names_to_species[synonym] = species
    for reference_name in reference_assemblies:
        if reference_name in _reference_names_to_species:
            raise ValueError("Can't use reference '%s' for both %s and %s" % (
                reference_name,
                species,
                _reference_names_to_species[reference_name]))
        _reference_names_to_species[reference_name] = species
    return species


def normalize_species_name(name):
    """
    If species name was "Homo sapiens" then replace spaces with underscores
    and return "homo_sapiens". Also replace common names like "human" with
    "homo_sapiens".
    """
    lower_name = name.lower().strip()

    # if given a common name such as "human", look up its latin equivalent
    if lower_name in _common_names_to_species:
        return _common_names_to_species[lower_name].latin_name

    return lower_name.replace(" ", "_")

def normalize_reference_name(name):
    """
    Search the dictionary of species-specific references to find a reference
    name that matches aside from capitalization.

    If no matching reference is found, raise an exception.
    """
    lower_name = name.strip().lower()
    for reference in _reference_names_to_species.keys():
        if reference.lower() == lower_name:
            return reference
    raise ValueError("Reference genome '%s' not found" % name)

def find_species_by_name(species_name):
    latin_name = normalize_species_name(species_name)
    if latin_name not in _latin_names_to_species:
        raise ValueError("Species not found: %s" % species_name)
    return _latin_names_to_species[latin_name]

def find_species_by_reference(reference_name):
    return _reference_names_to_species[normalize_reference_name(reference_name)]

def which_reference(species_name, ensembl_release):
    return find_species_by_name(species_name).which_reference(ensembl_release)

def max_ensembl_release(reference_name):
    species = find_species_by_reference(reference_name)
    (_, max_release) = species.reference_assemblies[reference_name]
    return max_release

def check_species_object(species_name_or_object):
    """
    Helper for validating user supplied species names or objects.
    """
    if isinstance(species_name_or_object, Species):
        return species_name_or_object
    elif isinstance(species_name_or_object, str):
        return find_species_by_name(species_name_or_object)
    else:
        raise ValueError("Unexpected type for species: %s : %s" % (
            species_name_or_object, type(species_name_or_object)))

human = add_species(
    latin_name="homo_sapiens",
    synonyms=["human"],
    reference_assemblies={
        "GRCh38": (76, MAX_ENSEMBL_RELEASE),
        "GRCh37": (55, 75),
        "NCBI36": (54, 54),
    })

mouse = add_species(
    latin_name="mus_musculus",
    synonyms=["mouse", "house mouse"],
    # TODO: fix release range
    # The range of releases given for GRCm38  isn't right and we don't have
    # ranges for other assemblies.
    reference_assemblies={
        "GRCm38": (78, MAX_ENSEMBL_RELEASE)
    })

"""
TODO: add all these species

_latin_to_common_names = {
    "ailuropoda_melanoleuca": ["panda"],
    "anas_platyrhynchos": ["mallard", "duck"],
    "anolis_carolinensis": ["carolina anole"],
    "astyanax_mexicanus": ["mexican tetra"],
    "bos_taurus": ["cattle", "cow"],
    "caenorhabditis_elegans": ["c. elegans", "nematode"],
    "callithrix_jacchus": ["marmoset"],
    "canis_familiaris": ["dog"],
    "cavia_porcellus": ["guinea pig"],
    "chlorocebus_sabaeus": ["green monkey", "sabaeus monkey"],
    "choloepus_hoffmanni": ["hoffman's two-toed sloth"],
    "ciona_intestinalis": ["vase tunicate"],
    "ciona_savignyi": [
        "pacific transparent sea squirt",
        "solitary sea squirt"],
    "danio_rerio": ["zebrafish"],
    "dasypus_novemcinctus": ["nine-banded armadillo"],
    "dipodomys_ordii": ["ord's kangaroo rat"],
    "drosophila_melanogaster": [
        "fruit fly",
        "vinegar fly"],
    "echinops_telfairi": ["lesser hedgehog tenrec"],
    "equus_caballus": ["horse"],
    "erinaceus_europaeus": [
        "european hedgehog",
        "hedgehog"],
    "felis_catus": ["cat", "domestic cat"],
    "ficedula_albicollis": ["collared flycatcher"],
    "gadus_morhua": ["atlnatic cod"],
    "gallus_gallus": ["chicken", "domesticated fowl"],
    "gasterosteus_aculeatus": ["three-spined stickleback"],
    "gorilla_gorilla": ["gorilla"],
    human_latin_name: ["human"],
    "ictidomys_tridecemlineatus": [
        "thirteen-lined ground squirrel",
        "striped gopher",
        "leopard ground squirrel"
        "squinney"],
    "latimeria_chalumnae": [
        "west indian ocean coelacanth",
        "african coelacanth"],
    "lepisosteus_oculatus": ["spotted gar"],
    "loxodonta_africana": ["african bush elephant"],
    "macaca_mulatta": ["rhesus macaque"],
    "macropus_eugenii": [
        "tammar wallaby",
        "dama wallaby",
        "darma wallaby"],
    "meleagris_gallopavo": ["wild turkey"],
    "microcebus_murinus": ["gray mouse lemur"],
    "monodelphis_domestica": ["gray short-tailed opossum"],
    mouse_latin_name: ["house mouse", "mouse"],
    "mustela_putorius_furo": ["domestic ferret"],
    "myotis_lucifugus": ["little brown bat"],
    "nomascus_leucogenys": ["northern white-cheeked gibbon"],
    "ochotona_princeps": ["american pika"],
    "oreochromis_niloticus": ["nile tilapia"],
    "ornithorhynchus_anatinus": ["platypus"],
    "oryctolagus_cuniculus": ["european rabbit", "rabbit"],
    "oryzias_latipes": ["japanese rice fish", "medaka"],
    "otolemur_garnettii": [
        "northern greater galago",
        "small-eared greater galago"],
    "ovis_aries": ["sheep"],
    "pan_troglodytes": ["chimpanzee"],
    "papio_anubis": ["olive baboon"],
    "pelodiscus_sinensis": ["chinese softshell turtle"],
    "petromyzon_marinus": ["sea lamprey"],
    "poecilia_formosa": ["amazon molly"],
    "pongo_abelii": ["sumatran orangutan"],
    "procavia_capensis": ["rock hyrax"],
    "pteropus_vampyrus": [
        "large flying fox",
        "large fruit bat",
        "kalang",
        "kalong"],
    "rattus_norvegicus": ["brown rat", "rat", "norwegian rat"],
    "saccharomyces_cerevisiae": ["yeast"],
    "sarcophilus_harrisii": ["tasmanian devil"],
    "sorex_araneus": ["shrew"],
    "sus_scrofa": ["wild boar"],
    "taeniopygia_guttata": ["zebra finch"],
    "takifugu_rubripes": ["japanese puffer", "tiger puffer", "torafugu"],
    "tarsius_syrichta": ["philippine tarsier", "mawmag", "mamag"],
    "tetraodon_nigroviridis": ["green spotted puffer"],
    "tupaia_belangeri": ["northern treeshrew"],
    "tursiops_truncatus": ["bottlenose dolphin"],
    "vicugna_pacos": ["whale"],
    "xenopus_tropicalis": ["western clawed frog"],
    "xiphophorus_maculatus": ["southern platyfish"],
}
"""



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cache and serializing the results of expensive computations. Used in pyensembl
primarily to cache the heavy-weight parsing of GTF files and various
filtering operations on Ensembl entries.

A piece of data is returned from one of three sources:
1) Cache cold. Run the user-supplied compute_fn.
2) Cache warm on disk. Parse or unpickle the serialized result into memory.
3) Cache warm in memory. Return cached object.
"""
from __future__ import print_function, division, absolute_import

import logging
from os import remove, stat
from os.path import exists

import pandas as pd

from .common import load_pickle, dump_pickle

class MemoryCache(object):
    def __init__(self):
        self._memory_cache = {}

    def is_empty(self, filename):
        return stat(filename).st_size == 0

    def delete_file(self, path):
        if exists(path):
            logging.info("Deleting cached file %s" % path)
            remove(path)

    def remove_from_cache(self, key):
        if key in self._memory_cache:
            del self._memory_cache[key]
        self.delete_file(key)

    def clear_cached_objects(self):
        for key in self._memory_cache.keys():
            self.delete_file(key)
        self._memory_cache.clear()

    def _read_csv(self, csv_path):
        print("Reading Dataframe from %s" % csv_path)
        df = pd.read_csv(csv_path)
        if 'seqname' in df:
            # by default, Pandas will infer the type as int,
            # then switch to str when it hits non-numerical
            # chromosomes. Make sure whole column has the same type
            df['seqname'] = df['seqname'].map(str)
        return df

    def _write_csv(self, df, csv_path, chunksize=10**5):
        """
        Parameters
        ----------
        df : pandas.DataFrame

        csv_path : str

        chunksize : int
            Number of rows to write at a time. Helps to limit memory
            consumption while writing a CSV.
        """
        print("Saving DataFrame to %s" % csv_path)
        df.to_csv(csv_path, index=False, chunksize=chunksize)

    def cached_dataframe(self, csv_path, compute_fn):
        """
        If a CSV path is in the _memory_cache, then return that cached value.

        If we've already saved the DataFrame as a CSV then load it.

        Otherwise run the provided `compute_fn`, and store its result
        in memory and and save it as a CSV.
        """
        if not csv_path.endswith(".csv"):
            raise ValueError("Invalid path '%s', must be a CSV file" % csv_path)

        if csv_path in self._memory_cache:
            return self._memory_cache[csv_path]

        if exists(csv_path) and not self.is_empty(csv_path):
            df = self._read_csv(csv_path)
        else:
            df = compute_fn()
            if not isinstance(df, pd.DataFrame):
                raise TypeError(
                    "Expected compute_fn to return DataFrame, got %s : %s" % (
                        df, type(df)))
            self._write_csv(df, csv_path)
        self._memory_cache[csv_path] = df
        return df

    def cached_object(self, path, compute_fn):
        """
        If `cached_object` has already been called for a value of `path` in this
        running Python instance, then it should have a cached value in the
         _memory_cache; return that value.

        If this function was never called before with a particular value of
        `path`, then call compute_fn, and pickle it to `path`.

        If `path` already exists, unpickle it and store that value in
        _memory_cache.
        """
        if path in self._memory_cache:
            return self._memory_cache[path]

        if exists(path) and not self.is_empty(path):
            obj = load_pickle(path)
        else:
            obj = compute_fn()
            dump_pickle(obj, path)
        self._memory_cache[path] = obj
        return obj



from __future__ import absolute_import

from pyensembl import Genome, GTF
from nose.tools import eq_

from .common import TemporaryDirectory
from .data import data_path

UCSC_GENCODE_PATH = data_path("gencode.ucsc.small.gtf")
UCSC_REFSEQ_PATH = data_path("refseq.ucsc.small.gtf")

def test_ucsc_gencode_gtf():
    with TemporaryDirectory() as tmpdir:
        gtf = GTF(
            UCSC_GENCODE_PATH,
            cache_directory_path=tmpdir)
        df = gtf.dataframe(save_to_disk=False)
        exons = df[df["feature"] == "exon"]
        # expect 12 exons from the dataframe
        assert len(exons) == 12, \
            "Expected 12 exons, got %d: %s" % (len(exons), exons)
        df2 = gtf.dataframe(save_to_disk=True)
        assert len(df) == len(df2), "Got different length DataFrame"
        assert list(df.columns) == list(df2.columns)


def test_ucsc_gencode_genome():
    """
    Testing with a small GENCODE GTF file downloaded from
    http://genome.ucsc.edu/cgi-bin/hgTables
    """
    with TemporaryDirectory() as tmpdir:
        genome = Genome(
            reference_name="GRCh38",
            annotation_name="ucsc_test",
            gtf_path_or_url=UCSC_GENCODE_PATH,
            cache_directory_path=tmpdir)
        genome.index()
        genes = genome.genes()
        for gene in genes:
            assert gene.id, \
                "Gene with missing ID in %s" % (genome.gtf.dataframe(),)
        assert len(genes) == 7, \
            "Expected 7 genes, got %d: %s" % (
                len(genes), genes)
        transcripts = genome.transcripts()
        for transcript in transcripts:
            assert transcript.id, \
                "Transcript with missing ID in %s" % (genome.gtf.dataframe(),)
        assert len(transcripts) == 7, \
            "Expected 7 transcripts, got %d: %s" % (
                len(transcripts), transcripts)

        gene_uc001aak4 = genome.gene_by_id("uc001aak.4")
        eq_(gene_uc001aak4.id, "uc001aak.4")
        eq_(gene_uc001aak4.name, None)
        eq_(gene_uc001aak4.biotype, None)

        gene_1_17369 = genome.genes_at_locus(1, 17369)
        eq_(gene_1_17369[0].id, "uc031tla.1")

        transcript_1_30564 = genome.transcripts_at_locus(1, 30564)
        eq_(transcript_1_30564[0].id, "uc057aty.1")

def test_ucsc_refseq_gtf():
    """
    Test GTF object with a small RefSeq GTF file downloaded from
    http://genome.ucsc.edu/cgi-bin/hgTables
    """
    with TemporaryDirectory() as tmpdir:
        gtf = GTF(
            UCSC_REFSEQ_PATH,
            cache_directory_path=tmpdir)
        df = gtf.dataframe(save_to_disk=False)
        exons = df[df["feature"] == "exon"]
        # expect 16 exons from the GTF
        assert len(exons) == 16, \
            "Expected 16 exons, got %d: %s" % (
                len(exons), exons)
        df2 = gtf.dataframe(save_to_disk=True)
        assert len(df) == len(df2), "Got different length DataFrame"
        assert list(df.columns) == list(df2.columns)

def test_ucsc_refseq_genome():
    """
    Test Genome object with a small RefSeq GTF file downloaded from
    http://genome.ucsc.edu/cgi-bin/hgTables
    """
    with TemporaryDirectory() as tmpdir:
        genome = Genome(
            reference_name="GRCh38",
            annotation_name="ucsc_test",
            gtf_path_or_url=UCSC_REFSEQ_PATH,
            cache_directory_path=tmpdir)
        genome.index()
        genes = genome.genes()
        for gene in genes:
            assert gene.id, \
                "Gene with missing ID in %s" % (genome.gtf.dataframe(),)
        assert len(genes) == 2, \
            "Expected 2 genes, got %d: %s" % (
                len(genes), genes)
        transcripts = genome.transcripts()
        for transcript in transcripts:
            assert transcript.id, \
                 "Transcript with missing ID in %s" % (genome.gtf.dataframe(),)
        assert len(transcripts) == 2, \
            "Expected 2 transcripts, got %d: %s" % (
                len(transcripts), transcripts)
        genes_at_locus = genome.genes_at_locus(1, 67092176)
        assert len(genes_at_locus) == 2, \
            "Expected 2 genes at locus chr1:67092176, got %d: %s" % (
                len(genes_at_locus), genes_at_locus)
        ids = set([gene.id for gene in genes_at_locus])
        eq_(set(["NM_001276352", "NR_075077"]), ids)



from __future__ import absolute_import

from pyensembl import find_nearest_locus
from nose.tools import eq_
from .common import test_ensembl_releases

@test_ensembl_releases()
def test_find_nearest_BRAF_exon(ensembl):
    braf = ensembl.genes_by_name("BRAF")[0]
    braf_transcripts = braf.transcripts
    exons = braf_transcripts[0].exons
    for exon in exons:
        # immediately before exon
        result_before = find_nearest_locus(
            start=exon.start - 2,
            end=exon.start - 1,
            loci=exons)
        eq_(result_before, (1, exon))

        # overlapping with exon
        result_overlap = find_nearest_locus(
            start=exon.start - 2,
            end=exon.start + 1,
            loci=exons)
        eq_(result_overlap, (0, exon))

        # immediately after exon
        result_after = find_nearest_locus(
            start=exon.end + 1,
            end=exon.end + 2,
            loci=exons)
        eq_(result_after, (1, exon))

@test_ensembl_releases()
def test_find_nearest_BRAF_transcript(ensembl):
    braf_transcript = ensembl.transcripts_by_name("BRAF-001")[0]
    egfr_transcript = ensembl.transcripts_by_name("EGFR-001")[0]
    transcripts = [braf_transcript, egfr_transcript]
    for transcript in transcripts:
        # immediately before transcript
        result_before = find_nearest_locus(
            start=transcript.start - 2,
            end=transcript.start - 1,
            loci=transcripts)
        eq_(result_before, (1, transcript))

        # overlapping with transcript
        result_overlap = find_nearest_locus(
            start=transcript.start - 2,
            end=transcript.start + 1,
            loci=transcripts)
        eq_(result_overlap, (0, transcript))

        # immediately after transcript
        # may overlap with other transcripts
        result_after = find_nearest_locus(
            start=transcript.end + 1,
            end=transcript.end + 2,
            loci=transcripts)
        eq_(result_after, (1, transcript))



"""
Check that pyensembl returns correct exon objects for exon IDs
and loci. Make sure the information on the exon object matches
the expected gene ID and location.
"""
from __future__ import absolute_import

from pyensembl import ensembl_grch38 as ensembl

def test_exon_object_by_id():
    """
    test_exon_object_by_id : check properties of exon 4 of CTNNB1 when looked
    up by ID in Ensembl 77.
    """
    exon = ensembl.exon_by_id("ENSE00003464041")
    assert exon.gene_name == "CTNNB1", \
        "Unexpected gene name: %s" % exon.gene_name
    assert exon.contig == "3", exon.contig
    assert exon.strand == "+"
    assert exon.on_forward_strand
    assert exon.on_positive_strand
    assert exon.start == 41224526, "Unexpected exon start: %s" % exon.start
    assert exon.end == 41224753, "Unexpected exon end: %s" % exon.end
    assert exon.length == len(exon) == 228

def test_exon_object_by_id_on_negative_strand():
    """
    test_exon_object_by_id : check properties of exon 1 from CXCR3 when looked
    up by ID in Ensembl 77.
    """
    exon = ensembl.exon_by_id("ENSE00001817013")
    assert exon.gene_name == "CXCR3", \
        "Unexpected gene name: %s" % exon.gene_name
    assert exon.contig == "X", exon.contig
    assert exon.strand == "-"
    assert exon.on_backward_strand
    assert exon.on_negative_strand
    assert exon.start == 71618438, "Unexpected exon start: %s" % exon.start
    assert exon.end == 71618517, "Unexpected exon end: %s" % exon.end
    assert exon.length == len(exon) == 80


def test_exon_object_at_locus():
    """
    test_exon_object_at_locus : check properties of exon 4 of CTNNB1 when looked
    up by its location on the forward strand of chr3
    """
    exons = ensembl.exons_at_locus(3, 41224526, strand="+")
    for exon in exons:
        assert exon.gene_name == "CTNNB1", exon.transcript_name
        assert exon.contig == "3", exon.contig
        assert exon.strand == "+"
        assert exon.on_forward_strand
        assert exon.on_positive_strand
        assert exon.start <= 41224526, "Unexpected exon start: %s" % exon.start
        assert exon.end >= 41224526, "Unexpected exon end: %s" % exon.end

def test_exon_object_at_locus_on_negative_strand():
    """
    test_exon_object_at_locus : check properties of exon 1 of CXCR3 when looked
    up by its location on the negative strand of chrX
    """
    exons = ensembl.exons_at_locus("chrX", 71618517, strand="-")
    for exon in exons:
        assert exon.gene_name == "CXCR3", exon.transcript_name
        assert exon.contig == "X", exon.contig
        assert exon.strand == "-"
        assert exon.on_backward_strand
        assert exon.on_negative_strand
        assert exon.start <= 71618517, "Unexpected exon start: %s" % exon.start
        assert exon.end >= 71618517, "Unexpected exon end: %s" % exon.end



"""
Tests for simple properties of an EnsemblRelease object which don't
require database lookups.
"""

from __future__ import absolute_import

from nose.tools import eq_
from pyensembl import EnsemblRelease

def test_reference_name():
    eq_(EnsemblRelease(release=54).reference_name, "NCBI36")
    eq_(EnsemblRelease(release=74).reference_name, "GRCh37")
    eq_(EnsemblRelease(release=75).reference_name, "GRCh37")
    eq_(EnsemblRelease(release=78).reference_name, "GRCh38")
    eq_(EnsemblRelease(release=79).reference_name, "GRCh38")


# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
from os.path import split, abspath, join, exists, splitext
import pandas as pd

from typechecks import require_string
from gtfparse import read_gtf_as_dataframe, create_missing_features

from .locus import normalize_chromosome, normalize_strand
from .memory_cache import MemoryCache

class GTF(object):
    """
    Parse a GTF gene annotation file from a given local path.
    Represent its contents as a Pandas DataFrame (optionally filtered
    by locus, column, contig, &c).
    """
    def __init__(self, gtf_path, cache_directory_path=None):

        if not (gtf_path.endswith(".gtf") or gtf_path.endswith(".gtf.gz")):
            raise ValueError("Wrong extension for GTF file: %s" % (gtf_path,))
        self.gtf_path = abspath(gtf_path)

        if not exists(self.gtf_path):
            raise ValueError("GTF file %s does not exist" % (self.gtf_path,))

        self.gtf_directory_path, self.gtf_filename = split(self.gtf_path)
        self.gtf_base_filename = splitext(self.gtf_filename)[0]

        # if cache directory isn't given then put cached files
        # alongside the GTF
        if cache_directory_path:
            self.cache_directory_path = cache_directory_path
        else:
            self.cache_directory_path = self.gtf_directory_path

        self.memory_cache = MemoryCache()

        # lazily load DataFrame of all GTF entries if necessary
        # for database construction
        self._dataframes = {}

    def __eq__(self, other):
        return (
            isinstance(other, GTF) and
            other.gtf_path == self.gtf_path)

    def __hash__(self):
        return hash(self.gtf_path)

    def clear_cache(self):
        # clear any dataframes we constructed
        self._dataframes.clear()

        # clear cached dataframes loaded from CSV
        self.memory_cache.clear_cached_objects()

    def data_subset_path(
            self,
            contig=None,
            feature=None,
            column=None,
            strand=None,
            distinct=False,
            extension=".csv"):
        """
        Path to cached file for storing materialized views of the genomic data.
        Typically this is a CSV file, the filename reflects which filters have
        been applied to the entries of the database.

        Parameters:

        contig : str, optional
            Path for subset of data restricted to given contig

        feature : str, optional
            Path for subset of data restrict to given feature

        column : str, optional
            Restrict to single column

        strand : str, optional
            Positive ("+") or negative ("-") DNA strand. Default = either.

        distinct : bool, optional
            Only keep unique values (default=False)
        """
        csv_filename = self.gtf_base_filename + ".expanded"
        if contig:
            contig = normalize_chromosome(contig)
            csv_filename += ".contig.%s" % (contig,)
        if feature:
            csv_filename += ".feature.%s" % (feature,)
        if column:
            csv_filename += ".column.%s" % (column,)
        if strand:
            if strand == "+":
                strand_string = "positive"
            elif strand == "-":
                strand_string = "negative"
            else:
                raise ValueError("Invalid strand value: %s" % strand)
            csv_filename += ".strand.%s" % strand_string
        if distinct:
            csv_filename += ".distinct"
        csv_filename += extension
        return join(self.cache_directory_path, csv_filename)

    def _load_full_dataframe_cached(self):
        """
        Loads full dataframe from cached CSV or constructs it from GTF
        """
        return self.memory_cache.cached_dataframe(
            csv_path=self.data_subset_path(),
            compute_fn=self._load_full_dataframe_from_gtf)

    def _load_full_dataframe_from_gtf(self):
        """
        Parse this genome source's GTF file and load it as a Pandas DataFrame
        """
        df = read_gtf_as_dataframe(
            self.gtf_path,
            column_converters={
                "seqname": normalize_chromosome,
                "strand": normalize_strand,
            },
            infer_biotype_column=True)

        features = set(df["feature"])
        column_names = set(df.keys())

        # older Ensembl releases don't have "gene" or "transcript"
        # features, so fill in those rows if they're missing
        if "gene" not in features:
            # if we have to reconstruct gene feature rows then
            # fill in values for 'gene_name' and 'gene_biotype'
            # but only if they're actually present in the GTF
            df = create_missing_features(
                dataframe=df,
                unique_keys={"gene": "gene_id"},
                extra_columns={
                    "gene": {
                        "gene_name",
                        "gene_biotype"
                    }.intersection(column_names),
                },
                missing_value="")

        if "transcript" not in features:
            df = create_missing_features(
                dataframe=df,
                unique_keys={"transcript": "transcript_id"},
                extra_columns={
                    "transcript": {
                        "gene_id",
                        "gene_name",
                        "gene_biotype",
                        "transcript_name",
                        "transcript_biotype",
                        "protein_id",
                    }.intersection(column_names)
                },
                missing_value="")
        return df

    def dataframe(
            self,
            contig=None,
            feature=None,
            strand=None,
            save_to_disk=False):
        """
        Load genome entries as a DataFrame, optionally restricted to
        particular contig or feature type.
        """
        if contig:
            contig = normalize_chromosome(contig)

        if strand:
            strand = normalize_strand(strand)

        if feature is not None:
            require_string(feature, "feature")

        key = (contig, feature, strand)

        if key not in self._dataframes:
            def _construct_df():
                full_df = self._load_full_dataframe_cached()

                assert len(full_df) > 0, \
                    "Dataframe representation of genomic database empty!"

                # rename since we're going to be filtering the entries but
                # may still want to access the full dataset
                df = full_df
                if contig:
                    df = df[df["seqname"] == contig]
                    if len(df) == 0:
                        raise ValueError("Contig not found: %s" % (contig,))

                if feature:
                    df = df[df["feature"] == feature]
                    if len(df) == 0:
                        # check to make sure feature was somewhere in
                        # the full dataset before returning an empty dataframe
                        features = full_df["feature"].unique()
                        if feature not in features:
                            raise ValueError(
                                "Feature not found: %s" % (feature,))
                if strand:
                    df = df[df["strand"] == strand]

                return df
            if save_to_disk:
                csv_path = self.data_subset_path(
                    contig=contig,
                    feature=feature,
                    strand=strand,
                    distinct=False)
                df = self.memory_cache.cached_dataframe(
                    csv_path=csv_path,
                    compute_fn=_construct_df)
            else:
                df = _construct_df()
            self._dataframes[key] = df

        return self._dataframes[key]

    def dataframe_column_at_locus(
            self,
            column_name,
            contig,
            start,
            end=None,
            offset=None,
            strand=None):
        """
        Subset of entries which overlap an inclusive range of loci
        """
        if end is None and offset is None:
            end = start
        elif offset is None:
            end = start + offset - 1

        df_contig = self.dataframe(contig=contig, strand=strand)

        assert column_name in df_contig, \
            "Unknown genome property: %s" % column_name

        return self._slice_column(
            df_contig[column_name],
            df_contig.start,
            df_contig.end,
            start,
            end)

    @staticmethod
    def _slice_column(column_name_series, start_series, end_series,
                      start, end):
        """
        Given a Series of data, and two Series' representing start
        and end positions for the end, keep all data for which
        the start and end positions overlap the start and end
        parameters (inclusively).

        For example:
        column_name_series: ENSG00000223972, ENSG00000000003, etc.
        start_series: 11869, 100635558, etc.
        end_series: 14409, 100635252, etc.
        start = 100000000
        end = 110000000

        Returns only ENSG00000000003, since it overlaps
        [100000000, 110000000].
        """
        df = pd.DataFrame({
            'name': column_name_series,
            'start': start_series,
            'end': end_series
        })
        df = GTF._slice(df, 'start', 'end', start, end)
        return df.name

    @staticmethod
    def _slice(df, df_start_col_name, df_end_col_name, start, end):
        """
        Given a DataFrame, along with the names of columns in the
        DataFrame representing start and end positions, keep all
        data for which the start and end positions overlap the start
        and end parameters (inclusively).
        """
        # No overlap because the whole thing is before start
        mask_left = df[df_end_col_name] < start
        # No overlap because the whole thing is after end
        mask_right = df[df_start_col_name] > end
        df = df[~mask_left & ~mask_right]
        return df

    def dataframe_at_locus(
            self,
            contig,
            start,
            end=None,
            offset=None,
            strand=None):
        """
        Subset of entries which overlap an inclusive range of
        chromosomal positions
        """
        if end is None and offset is None:
            end = start
        elif offset is None:
            end = start + offset - 1

        df_contig = self.dataframe(contig=contig, strand=strand)

        # find genes whose start/end boundaries overlap with the position
        return GTF._slice(df_contig, df_contig.start.name,
                          df_contig.end.name, start, end)



from __future__ import absolute_import

from nose.tools import assert_raises
from pyensembl.download_cache import (
    DownloadCache,
    MissingLocalFile,
    MissingRemoteFile
)

download_cache = DownloadCache(
    reference_name="__test_reference",
    annotation_name="__test_annotation",
    copy_local_files_to_cache=False)

def test_download_cache_missing_local_file():
    # clear the cache
    download_cache.delete_cache_directory()
    with assert_raises(MissingLocalFile):
        download_cache.download_or_copy_if_necessary(
            path_or_url="test_file_doesn_not_exist.file")

def test_download_cache_missing_remote_file():
    # clear the cache
    download_cache.delete_cache_directory()
    with assert_raises(MissingRemoteFile):
        download_cache.download_or_copy_if_necessary(
            path_or_url="ftp://NOTAURL.NOTAURL.NOTAURL")



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from six.moves import intern
from typechecks import is_string, is_integer

# Manually memoizing here, since our simple common.memoize function has
# noticable overhead in this instance.
NORMALIZE_CHROMOSOME_CACHE = {}

def normalize_chromosome(c):
    try:
        return NORMALIZE_CHROMOSOME_CACHE[c]
    except KeyError:
        pass

    if not (is_string(c) or is_integer(c)):
        raise TypeError("Chromosome cannot be '%s' : %s" % (c, type(c)))

    result = str(c)
    if result == "0":
        raise ValueError("Chromosome name cannot be 0")
    elif result == "":
        raise ValueError("Chromosome name cannot be empty")

    # only strip off lowercase chr since some of the non-chromosomal
    # contigs start with "CHR"
    if result.startswith("chr"):
        result = result[3:]

    # standardize mitochondrial genome to be "MT"
    if result == "M":
        result = "MT"
    else:
        # just in case someone is being lazy, capitalize "X" and "Y"
        result = result.upper()
    # interning strings since the chromosome names probably get constructed
    # or parsed millions of times, can save memory in tight situations
    # (such as parsing GTF files)
    result = intern(result)
    NORMALIZE_CHROMOSOME_CACHE[c] = result
    return result

def normalize_strand(strand):
    if strand == "+" or strand == "-":
        return strand
    elif strand == 1:
        return "+"
    elif strand == -1:
        return "-"

    raise ValueError("Invalid strand: %s" % (strand,))

class Locus(object):
    """
    Base class for any entity which can be localized at a range of positions
    on a particular strand of a chromosome/contig.
    """

    def __init__(
            self,
            contig,
            start,
            end,
            strand):
        """
        contig : str
            Chromosome or other sequence name in the reference assembly

        start : int
            Start position of locus on the contig

        end : int
            Inclusive end position on the contig

        strand : str
            Should we read the locus forwards ('+') or backwards ('-')?
        """

        self.contig = normalize_chromosome(contig)
        self.strand = normalize_strand(strand)

        start = int(start)
        end = int(end)

        if start == 0:
            raise ValueError("Expected start > 0 (using base 1 coordinates)")
        elif end == 0:
            raise ValueError("Expected end > 0 (using base 1 coordinates)")

        if end < start:
            raise ValueError(
                "Expected start <= end, got start = %d, end = %d" % (
                    start, end))
        self.start = start
        self.end = end

    def __str__(self):
        return "Locus(contig=%s, start=%s, end=%s, strand=%s)" % (
            self.contig, self.start, self.end, self.strand)

    def __repr__(self):
        return str(self)

    def __len__(self):
        return self.end - self.start + 1

    def __eq__(self, other):
        return (
            isinstance(other, Locus) and
            self.contig == other.contig and
            self.start == other.start and
            self.end == other.end and
            self.strand == other.strand
        )

    def __hash__(self):
        return hash(str(self))

    @property
    def length(self):
        return self.end - self.start + 1

    def offset(self, position):
        """Offset of given position from stranded start of this locus.

        For example, if a Locus goes from 10..20 and is on the negative strand,
        then the offset of position 13 is 7, whereas if the Locus is on the
        positive strand, then the offset is 3.
        """
        if position > self.end or position < self.start:
            raise ValueError(
                "Position %d outside valid range %d..%d of %s" % (
                    position, self.start, self.end, self))
        elif self.on_forward_strand:
            return position - self.start
        else:
            return self.end - position

    def offset_range(self, start, end):
        """
        Database start/end entries are always ordered such that
        start < end. This makes computing a relative position (e.g. of a stop
        codon relative to its transcript) complicated since the "end"
        position of a backwards locus is actually earlir on the strand.
        This function correctly selects a start vs. end value depending
        on this locuses's strand and determines that position's offset from
        the earliest position in this locus.
        """
        assert start <= end, \
            "Locations should always have start < end, got start=%d, end=%d" % (
                start, end)

        if start < self.start or end > self.end:
            raise ValueError("Range (%d, %d) falls outside %s" % (
                start, end, self))

        if self.on_forward_strand:
            return (start - self.start, end - self.start)

        else:
            return (self.end - end, self.end - start)

    def on_contig(self, contig):
        return normalize_chromosome(contig) == self.contig

    def on_strand(self, strand):
        return normalize_strand(strand) == self.strand

    @property
    def on_forward_strand(self):
        return self.on_strand("+")

    @property
    def on_positive_strand(self):
        return self.on_forward_strand

    @property
    def on_backward_strand(self):
        return self.on_strand("-")

    @property
    def on_negative_strand(self):
        return self.on_backward_strand

    def can_overlap(self, contig, strand=None):
        """
        Is this locus on the same contig and (optionally) on the same strand?
        """
        return (
            self.on_contig(contig)
            and
            (strand is None or self.on_strand(strand)))

    def distance_to_interval(self, start, end):
        """
        Find the distance between intervals [start1, end1] and [start2, end2].
        If the intervals overlap then the distance is 0.
        """
        if self.start > end:
            # interval is before this exon
            return self.start - end
        elif self.end < start:
            # exon is before the interval
            return start - self.end
        else:
            return 0

    def distance_to_locus(self, other):
        if not self.can_overlap(other.contig, other.strand):
            # if two loci are on different contigs or strands,
            # can't compute a distance between them
            return float("inf")
        return self.distance_to_interval(other.start, other.end)

    def overlaps(self, contig, start, end, strand=None):
        """
        Does this locus overlap with a given range of positions?

        Since locus position ranges are inclusive, we should make sure
        that e.g. chr1:10-10 overlaps with chr1:10-10
        """
        return (
            self.can_overlap(contig, strand) and
            self.distance_to_interval(start, end) == 0)

    def overlaps_locus(self, other_locus):
        return self.overlaps(
            other_locus.contig,
            other_locus.start,
            other_locus.end,
            other_locus.strand)

    def contains(self, contig, start, end, strand=None):
        return (
            self.can_overlap(contig, strand) and
            start >= self.start and
            end <= self.end)

    def contains_locus(self, other_locus):
        return self.contains(
            other_locus.contig,
            other_locus.start,
            other_locus.end,
            other_locus.strand)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
from os import remove
from os.path import exists, abspath, split, join
import logging

from six.moves import cPickle as pickle
from skbio import read

from .common import require_ensembl_id, load_pickle, dump_pickle

class SequenceData(object):
    """
    Container for reference nucleotide and amino acid sequenes.
    """
    def __init__(
            self,
            fasta_path,
            require_ensembl_ids=False,
            sequence_type=str,
            cache_directory_path=None):
        self.fasta_path = abspath(fasta_path)
        self.fasta_directory_path, self.fasta_filename = split(self.fasta_path)
        if cache_directory_path:
            self.cache_directory_path = cache_directory_path
        else:
            self.cache_directory_path = self.fasta_directory_path
        if not exists(self.fasta_path):
            raise ValueError("Couldn't find FASTA file %s" % (self.fasta_path,))
        self.require_ensembl_ids = require_ensembl_ids
        self.sequence_type = sequence_type
        self.fasta_dictionary_filename = self.fasta_filename + ".pickle"
        self.fasta_dictionary_pickle_path = join(
            self.cache_directory_path,
            self.fasta_dictionary_filename)
        self._init_lazy_fields()

    def _init_lazy_fields(self):
        self._fasta_dictionary = None
        self._fasta_keys = None

    def clear_cache(self):
        self._init_lazy_fields()
        if exists(self.fasta_dictionary_pickle_path):
            remove(self.fasta_dictionary_pickle_path)

    def __str__(self):
        return "SequenceData(fasta_path=%s, require_ensembl_ids=%s)" % (
            self.fasta_path,
            self.require_ensembl_ids)

    def __repr__(self):
        return str(self)

    def __contains__(self, sequence_id):
        if self._fasta_keys is None:
            self._fasta_keys = set(self.fasta_dictionary.keys())
        return sequence_id in self._fasta_keys

    def __eq__(self, other):
        return (
            isinstance(other, SequenceData) and
            self.fasta_path == other.fasta_path)

    def __hash__(self):
        return hash(self.fasta_path)

    def _parse_fasta_dictionary(self):
        fasta_dictionary = {}
        sequence_type = self.sequence_type
        for seq_entry in read(self.fasta_path, format="fasta"):
            seq_id = seq_entry.metadata["id"]
            fasta_dictionary[seq_id] = sequence_type(seq_entry)
        return fasta_dictionary

    def _load_or_create_fasta_dictionary_pickle(self):
        if exists(self.fasta_dictionary_pickle_path):
            # try loading the cached file
            # but we'll fall back on recreating it if loading fails
            try:
                self._fasta_dictionary = load_pickle(
                    self.fasta_dictionary_pickle_path)
                return
            except (pickle.UnpicklingError, AttributeError):
                # catch either an UnpicklingError or an AttributeError
                # resulting from pickled objects refering to classes
                # that no longer exists
                logging.warn(
                    "Failed to load %s, attempting to read FASTA directly" % (
                        self.fasta_dictionary_pickle_path,))
        self._fasta_dictionary = self._parse_fasta_dictionary()
        dump_pickle(self._fasta_dictionary, self.fasta_dictionary_pickle_path)

    def index(self, overwrite=False):
        if overwrite:
            self.clear_cache()
        self._load_or_create_fasta_dictionary_pickle()

    @property
    def fasta_dictionary(self):
        if not self._fasta_dictionary:
            self._load_or_create_fasta_dictionary_pickle()
        return self._fasta_dictionary

    def get(self, sequence_id):
        """Get sequence associated with given ID or return None if missing"""
        # all Ensembl identifiers start with ENS e.g. ENST, ENSP, ENSE
        if self.require_ensembl_ids:
            require_ensembl_id(sequence_id)
        return self.fasta_dictionary.get(sequence_id)



# Copyright (c) 2014. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import os

from setuptools import setup

current_directory = os.path.dirname(__file__)
readme_filename = 'README.md'
readme_path = os.path.join(current_directory, readme_filename)

readme = ""
try:
    with open(readme_path, 'r') as f:
        readme = f.read()
except IOError as e:
    print(e)
    print("Failed to open %s" % readme_path)

try:
    import pypandoc
    readme = pypandoc.convert(readme, to='rst', format='md')
except ImportError as e:
    print(e)
    print("Failed to convert %s to reStructuredText", readme_filename)
    pass

if __name__ == '__main__':
    setup(
        name='pyensembl',
        version="0.8.4",
        description="Python interface to ensembl reference genome metadata",
        author="Alex Rubinsteyn",
        author_email="alex {dot} rubinsteyn {at} mssm {dot} edu",
        url="https://github.com/hammerlab/pyensembl",
        license="http://www.apache.org/licenses/LICENSE-2.0.html",
        entry_points={
            'console_scripts': [
                'pyensembl = pyensembl.shell:run'
            ],
        },
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Operating System :: OS Independent',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
        ],
        install_requires=[
            "typechecks>=0.0.2",
            "numpy>=1.7",
            "pandas>=0.15",
            "datacache>=0.4.16",
            "scikit-bio>=0.4.0",
            "memoized-property>=1.0.2",
            "six>=1.9.0",
            "gtfparse>=0.0.3",
        ],
        long_description=readme,
        packages=['pyensembl'],
    )



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import listdir, remove
from os.path import join, exists, split, abspath
from shutil import copy2, rmtree
import logging

import datacache

CACHE_BASE_SUBDIR = "pyensembl"

def cache_subdirectory(
        reference_name=None,
        annotation_name=None,
        annotation_version=None):
    """
    Which cache subdirectory to use for a given annotation database
    over a particular reference. All arguments can be omitted to just get
    the base subdirectory for all pyensembl cached datasets.
    """
    if reference_name is None:
        reference_name = ""
    if annotation_name is None:
        annotation_name = ""
    if annotation_version is None:
        annotation_version = ""
    reference_dir = join(CACHE_BASE_SUBDIR, reference_name)
    annotation_dir = "%s%s" % (annotation_name, annotation_version)
    return join(reference_dir, annotation_dir)

class MissingRemoteFile(Exception):
    def __init__(self, url):
        self.url = url

class MissingLocalFile(Exception):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return("MissingFile(%s)" % self.path)

class DownloadCache(object):
    """
    Downloads remote files to cache, optionally copies local files into cache,
    raises custom message if data is missing.
    """
    def __init__(
            self,
            reference_name,
            annotation_name,
            annotation_version=None,
            decompress_on_download=False,
            copy_local_files_to_cache=False,
            install_string_function=None,
            cache_directory_path=None):
        """
        Parameters
        ----------
        reference_name : str
            Name of reference genome

        annotation_name : str
            Name of annotation database

        annotation_version : str or int, optional
            Version or release of annotation database

        decompress_on_download : bool, optional
            If downloading a .fa.gz file, should we automatically expand it
            into a decompressed FASTA file?

        copy_local_files_to_cache : bool, optional
            If file is on the local file system, should we still copy it
            into the cache?

        install_string_function : fn, optional
            Function which returns an error message with
            install instructions. If not provided then the error tells the
            user what data is missing without install instructions.

        cache_directory_path : str, optional
            Where to place downloaded and temporary files, by default
            inferred from reference name, annotation name, annotation version,
            and the global cache directory determined by datacache.
        """

        self.reference_name = reference_name
        self.annotation_name = annotation_name
        self.annotation_version = annotation_version

        # using hidden member variable _cache_directory path since access to
        # to the visible cache_directory_path (no underscore!) is combined
        # with ensuring that the directpry actually exists
        if cache_directory_path:
            self._cache_directory_path = cache_directory_path
        else:
            self.cache_subdirectory = cache_subdirectory(
                reference_name=reference_name,
                annotation_name=annotation_name,
                annotation_version=annotation_version)

            self._cache_directory_path = datacache.get_data_dir(
                subdir=self.cache_subdirectory)

        self.decompress_on_download = decompress_on_download
        self.copy_local_files_to_cache = copy_local_files_to_cache
        self.install_string_function = install_string_function

    @property
    def cache_directory_path(self):
        datacache.ensure_dir(self._cache_directory_path)
        return self._cache_directory_path

    def _fields(self):
        """
        Fields used for hashing, string representation, equality comparison
        """
        return (
            ('reference_name', self.reference_name,),
            ('annotation_name', self.annotation_name),
            ('annotation_version', self.annotation_version),
            ('cache_directory_path', self.cache_directory_path),
            ('decompress_on_download', self.decompress_on_download),
            ('copy_local_files_to_cache', self.copy_local_files_to_cache)
        )

    def __eq__(self, other):
        return (
            other.__class__ is DownloadCache and
            self._fields() == other._fields()
        )

    def __hash__(self):
        return hash(self._fields())

    def __str__(self):
        fields_str = ", ".join("%s=%s" % (k, v) for (k, v) in self._fields())
        return "DownloadCache(%s)" % fields_str

    def __repr__(self):
        return str(self)

    def is_url_format(self, path_or_url):
        assert path_or_url, "Expected non-empty string for path_or_url"
        return "://" in path_or_url

    def _remove_compression_suffix_if_present(self, filename):
        """
        If the given filename ends in one of the compression suffixes that
        datacache knows how to deal with, remove the suffix (since we expect
        the result of downloading to be a decompressed file)
        """
        for ext in [".gz", ".gzip", ".zip"]:
            if filename.endswith(ext):
                return filename[:-len(ext)]
        return filename


    def cached_path(self, path_or_url):
        """
        When downloading remote files, the default behavior is to name local
        files the same as their remote counterparts.
        """
        assert path_or_url, "Expected non-empty string for path_or_url"
        remote_filename = split(path_or_url)[1]
        if self.is_url_format(path_or_url):
            # passing `decompress=False` since there is logic below
            # for stripping decompression extensions for both local
            # and remote files
            local_filename = datacache.build_local_filename(
                download_url=path_or_url,
                filename=remote_filename,
                decompress=False)
        else:
            local_filename = remote_filename

        # if we expect the download function to decompress this file then
        # we should use its name without the compression extension
        if self.decompress_on_download:
            local_filename = self._remove_compression_suffix_if_present(
                local_filename)

        if len(local_filename) == 0:
            raise ValueError("Can't determine local filename for %s" % (
                path_or_url,))

        return join(self.cache_directory_path, local_filename)

    def _download_if_necessary(self, url, download_if_missing, overwrite):
        """
        Return local cached path to a remote file, download it if necessary.
        """
        cached_path = self.cached_path(url)
        missing = not exists(cached_path)
        if (missing or overwrite) and download_if_missing:
            logging.info("Fetching %s from URL %s", cached_path, url)
            local_filename = split(cached_path)[1]
            datacache.download._download(
                filename=local_filename,
                full_path=cached_path,
                download_url=url)
        elif missing:
            raise MissingRemoteFile(url)
        return cached_path

    def _copy_if_necessary(self, local_path, overwrite):
        """
        Return cached path to local file, copying it to the cache if necessary.
        """
        local_path = abspath(local_path)
        if not exists(local_path):
            raise MissingLocalFile(local_path)
        elif not self.copy_local_files_to_cache:
            return local_path
        else:
            cached_path = self.cached_path(local_path)
            if exists(cached_path) and not overwrite:
                return cached_path
            copy2(local_path, cached_path)
            return cached_path

    def download_or_copy_if_necessary(
            self,
            path_or_url,
            download_if_missing=False,
            overwrite=False):
        """
        Download a remote file or copy
        Get the local path to a possibly remote file.

        Download if file is missing from the cache directory and
        `download_if_missing` is True. Download even if local file exists if
        both `download_if_missing` and `overwrite` are True.

        If the file is on the local file system then return its path, unless
        self.copy_local_to_cache is True, and then copy it to the cache first.

        Parameters
        ----------
        path_or_url : str

        download_if_missing : bool, optional
            Download files if missing from local cache

        overwrite : bool, optional
            Overwrite existing copy if it exists
        """
        assert path_or_url, "Expected non-empty string for path_or_url"
        if self.is_url_format(path_or_url):
            return self._download_if_necessary(
                path_or_url,
                download_if_missing,
                overwrite)
        else:
            return self._copy_if_necessary(path_or_url, overwrite)

    def _raise_missing_file_error(self, missing_urls_dict):
        missing_urls = list(missing_urls_dict.values())
        n_missing = len(missing_urls)
        assert n_missing > 0
        error_message = "Missing genome data file%s from %s." % (
            ("s", missing_urls) if n_missing > 1 else ("", missing_urls[0]))
        if self.install_string_function:
            install_string = self.install_string_function()
            error_message += " Run %s" % install_string
        raise ValueError(error_message)

    def local_path_or_install_error(
            self,
            field_name,
            path_or_url,
            download_if_missing=False,
            overwrite=False):
        try:
            return self.download_or_copy_if_necessary(
                path_or_url,
                download_if_missing=download_if_missing,
                overwrite=overwrite)
        except MissingRemoteFile:
            self._raise_missing_file_error({field_name: path_or_url})

    def delete_cached_files(self, prefixes=[], suffixes=[]):
        """
        Deletes any cached files matching the prefixes or suffixes given
        """
        for filename in listdir(self.cache_directory_path):
            delete = (
                any(filename.endswith(ext) for ext in suffixes) or
                any(filename.startswith(pre) for pre in prefixes))
            if delete:
                path = join(self.cache_directory_path, filename)
                print("Deleting %s" % path)
                remove(path)

    def delete_cache_directory(self):
        rmtree(self.cache_directory_path)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the EnsemblRelease class, which extends the Genome class
to be specific to (a particular release of) Ensembl.
"""

from .genome import Genome
from .ensembl_release_versions import check_release_number, MAX_ENSEMBL_RELEASE
from .species import check_species_object, human

from .ensembl_url_templates import (
    ENSEMBL_FTP_SERVER,
    make_gtf_url,
    make_fasta_url
)

class EnsemblRelease(Genome):
    """
    Bundles together the genomic annotation and sequence data associated with
    a particular release of the Ensembl database.
    """
    def __init__(self,
                 release=MAX_ENSEMBL_RELEASE,
                 species=human,
                 server=ENSEMBL_FTP_SERVER):
        self.release = check_release_number(release)
        self.species = check_species_object(species)
        self.server = server

        self.gtf_url = make_gtf_url(
            ensembl_release=self.release,
            species=species,
            server=server)
        self.transcript_fasta_url = make_fasta_url(
            ensembl_release=self.release,
            species=self.species.latin_name,
            sequence_type="cdna",
            server=server)
        self.protein_fasta_url = make_fasta_url(
            ensembl_release=self.release,
            species=self.species.latin_name,
            sequence_type="pep",
            server=server)

        self.reference_name = self.species.which_reference(self.release)

        Genome.__init__(
            self,
            reference_name=self.reference_name,
            annotation_name="ensembl",
            annotation_version=self.release,
            gtf_path_or_url=self.gtf_url,
            transcript_fasta_path_or_url=self.transcript_fasta_url,
            protein_fasta_path_or_url=self.protein_fasta_url,
            require_ensembl_ids=True)

    def install_string(self):
        return "pyensembl install --release %d --species %s" % (
            self.release,
            self.species.latin_name)

    def __str__(self):
        return "EnsemblRelease(release=%d, species='%s')" % (
            self.release,
            self.species.latin_name)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            other.__class__ is EnsemblRelease and
            self.release == other.release and
            self.species == other.species)

    def __hash__(self):
        return hash((self.release, self.species))



from __future__ import absolute_import

from glob import glob
from os.path import join
from pyensembl import GTF

from .data import data_path
from .common import TemporaryDirectory

gtf_path = data_path("mouse.ensembl.81.partial.ENSMUSG00000017167.gtf")

def test_gtf_object_path():
    """
    Make sure that GTF doesn't do anything funny under the hood
    (such as copying local files to some cached directory), all of that
    should be done at the level of Genome/DownloadCache.
    """
    gtf_object = GTF(gtf_path)
    assert gtf_path == gtf_object.gtf_path

def test_gtf_creates_csv_files_in_cache_dir():
    """
    Make sure that GTF obeys the cache_dir argument by creating all of its
    index files there.
    """
    with TemporaryDirectory() as tmpdir:
        gtf_object = GTF(gtf_path, cache_directory_path=tmpdir)
        csv_path = gtf_object.data_subset_path()
        assert tmpdir in csv_path, \
            "Expected CSV path %s to contain cache_dir" % csv_path
        search_pattern = join(tmpdir, "*mouse*")
        assert len(glob(search_pattern)) == 0, \
            "Temporary directory should be empty"
        # creating the dataframe should have the effect of triggering
        # GTF parsing and then saving the parsed results in a csv file
        gtf_object.dataframe()
        assert len(glob(search_pattern)) > 0, \
            "Expected GTF to save files in cache_dir"


from __future__ import absolute_import

from pyensembl.locus import normalize_chromosome, Locus

from nose.tools import assert_raises

def test_normalize_chromosome():
    assert normalize_chromosome("X") == "X"
    assert normalize_chromosome("chrX") == "X"
    assert normalize_chromosome("x") == "X"

    assert normalize_chromosome(1) == "1"
    assert normalize_chromosome("1") == "1"
    assert normalize_chromosome("chr1") == "1"

    assert normalize_chromosome("chrM") == "MT"
    assert normalize_chromosome("chrMT") == "MT"
    assert normalize_chromosome("M") == "MT"
    assert normalize_chromosome("MT") == "MT"

    with assert_raises(TypeError):
        normalize_chromosome({"a": "b"})

    with assert_raises(TypeError):
        normalize_chromosome([])

    with assert_raises(TypeError):
        normalize_chromosome(None)

    with assert_raises(ValueError):
        normalize_chromosome("")

    with assert_raises(ValueError):
        normalize_chromosome(0)

def test_locus_overlaps():
    locus = Locus("1", 10, 20, "+")
    assert locus.overlaps("1", 10, 20, "+")
    assert locus.overlaps("1", 10, 20)
    assert locus.overlaps("1", 5, 30)
    assert locus.overlaps("1", 15, 16)
    assert locus.overlaps("1", 15, 30)
    assert locus.overlaps("1", 5, 15)
    assert locus.overlaps("1", 10, 10)
    assert locus.overlaps("1", 20, 20)
    # before start
    assert not locus.overlaps(1, 9, 9)
    # after end
    assert not locus.overlaps(21, 30, 30)
    # wrong contig
    assert not locus.overlaps("2", 10, 20)
    # wrong strand
    assert not locus.overlaps("1", 10, 20, "-")

def test_locus_contains():
    locus = Locus("1", 10, 20, "+")
    assert locus.contains("1", 10, 20, "+")
    assert locus.contains("1", 10, 20)
    assert locus.contains("1", 15, 16)
    assert locus.contains("1", 10, 10)
    assert locus.contains("1", 20, 20)

    # before start and after end
    assert not locus.contains("1", 5, 30)

    # before start
    assert not locus.contains("1", 1, 9)
    assert not locus.contains("1", 5, 15)

    # after end
    assert not locus.contains("1", 21, 30)
    assert not locus.contains("1", 15, 30)

    # wrong contig
    assert not locus.contains("2", 10, 20)

    # wrong strand
    assert not locus.contains("1", 10, 20, "-")

def test_position_offset():
    forward_locus = Locus("1", 10, 20, "+")
    assert forward_locus.offset(10) == 0
    assert forward_locus.offset(15) == 5
    assert forward_locus.offset(19) == 9
    assert forward_locus.offset(20) == 10

    negative_locus = Locus("1", 10, 20, "-")
    assert negative_locus.offset(10) == 10
    assert negative_locus.offset(15) == 5
    assert negative_locus.offset(19) == 1
    assert negative_locus.offset(20) == 0

    # don't allow negative offsets
    with assert_raises(ValueError):
        forward_locus.offset(9)

    # don't allow negative offsets
    with assert_raises(ValueError):
        negative_locus.offset(9)

    # don't allow offset past the end of the locus
    with assert_raises(ValueError):
        forward_locus.offset(21)

    # don't allow offset past the end of the locus
    with assert_raises(ValueError):
        negative_locus.offset(21)


def test_range_offset():
    forward_locus = Locus("1", 10, 20, "+")
    assert forward_locus.offset_range(10, 20) == (0, 10)
    assert forward_locus.offset_range(11, 14) == (1, 4)
    assert forward_locus.offset_range(20, 20) == (10, 10)

    negative_locus = Locus("1", 10, 20, "-")
    assert negative_locus.offset_range(10, 20) == (0, 10)
    assert negative_locus.offset_range(11, 14) == (6, 9)
    assert negative_locus.offset_range(20, 20) == (0, 0)

    # start shouldn't be larger than end
    with assert_raises(AssertionError):
        forward_locus.offset_range(21, 20)

    # start shouldn't be larger than end
    with assert_raises(AssertionError):
        negative_locus.offset_range(21, 20)

    # don't allow negative offsets
    with assert_raises(ValueError):
        forward_locus.offset_range(9, 10)

    # don't allow negative offsets
    with assert_raises(ValueError):
        forward_locus.offset_range(9, 10)

    # don't allow negative offsets
    with assert_raises(ValueError):
        negative_locus.offset_range(9, 10)

def test_locus_distance():
    locus_chr1_10_20_pos = Locus("1", 10, 20, "+")
    locus_chr1_21_25_pos = Locus("1", 21, 25, "+")
    locus_chr2_21_25_pos = Locus("2", 21, 25, "+")
    locus_chr1_21_25_neg = Locus("1", 21, 25, "-")
    assert locus_chr1_10_20_pos.distance_to_locus(locus_chr1_21_25_pos) == 1
    assert locus_chr1_21_25_pos.distance_to_locus(locus_chr1_10_20_pos) == 1
    inf = float("inf")
    assert locus_chr1_10_20_pos.distance_to_locus(locus_chr2_21_25_pos) == inf
    assert locus_chr1_10_20_pos.distance_to_locus(locus_chr1_21_25_neg) == inf



from __future__ import print_function, absolute_import

from pyensembl import ensembl_grch38 as ensembl

from tinytimer import benchmark

contigs = [str(i + 1) for i in range(22)] + ["X", "Y"]

def make_repeat_lookup_fn(lookup_fn, n_positions):
    """
    Make a thunk which calls the lookup_fn at a number of loci
    for each human chromosome (excluding MT).
    """
    def repeat_lookup_fn():
        for contig in contigs:
            for position in [10 ** 6 + i * 10 ** 6 for i in range(n_positions)]:
                lookup_fn(contig, position)
    return repeat_lookup_fn

def run_benchmark(lookup_fn, n_positions_per_contig=20, time_limit=60.0):
    """
    Take a lookup functions (such as EnsemblRelease.genes_at_locus) and
    time how long it takes across multiple loci.
    """
    repeat_lookup_fn = make_repeat_lookup_fn(lookup_fn, n_positions_per_contig)
    n_loci = n_positions_per_contig * len(contigs)
    name = lookup_fn.__name__
    average_time = benchmark(
        repeat_lookup_fn,
        name="%s for %d loci" % (name, n_loci))
    print("-- %s : %0.4fs" % (name, average_time))
    assert average_time < time_limit, \
        "%s took too long for %s loci: %0.4fs" % (name, n_loci, average_time)
    return average_time

def test_timing_genes_at_locus():
    run_benchmark(ensembl.genes_at_locus)

def test_timing_transcripts_at_locus():
    run_benchmark(ensembl.transcripts_at_locus)

def test_timing_exons_at_locus():
    run_benchmark(ensembl.exons_at_locus)

def test_timing_transcript_sequences_at_locus():
    def transcript_sequences_at_locus(contig, position):
        sequences = []
        for transcript in ensembl.transcripts_at_locus(contig, position):
            sequences.append(transcript.sequence)
        return sequences
    run_benchmark(transcript_sequences_at_locus)

def test_timing_transcript_coding_sequences_at_locus():
    def transcript_coding_sequences_at_locus(contig, position):
        sequences = []
        for transcript in ensembl.transcripts_at_locus(contig, position):
            if transcript.sequence and transcript.complete:
                sequences.append(transcript.coding_sequence)
        return sequences
    run_benchmark(transcript_coding_sequences_at_locus)

def run_all_benchmarks():
    import types

    # run all local test functions to see their timings printed
    global_variables = globals()
    for variable_name in global_variables:
        if "test_" in variable_name:
            f = global_variables[variable_name]
            if isinstance(f, types.FunctionType):
                f()

if __name__ == "__main__":
    run_all_benchmarks()
